# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
from typing import List, Literal, Optional, TypedDict, Union

from typing_extensions import NotRequired

from .automod import (
    AutoModAction,
    AutoModEventType,
    AutoModRule,
    AutoModTriggerMetadata,
    AutoModTriggerType,
)
from .channel import ChannelType, PermissionOverwrite, VideoQualityMode
from .guild import (
    DefaultMessageNotificationLevel,
    ExplicitContentFilterLevel,
    MFALevel,
    VerificationLevel,
)
from .guild_scheduled_event import GuildScheduledEvent
from .integration import IntegrationExpireBehavior, PartialIntegration
from .interactions import ApplicationCommand, ApplicationCommandPermissions
from .role import Role
from .snowflake import Snowflake
from .threads import Thread
from .user import User
from .webhook import Webhook

AuditLogEvent = Literal[
    1,
    10,
    11,
    12,
    13,
    14,
    15,
    20,
    21,
    22,
    23,
    24,
    25,
    26,
    27,
    28,
    30,
    31,
    32,
    40,
    41,
    42,
    50,
    51,
    52,
    60,
    61,
    62,
    72,
    73,
    74,
    75,
    80,
    81,
    82,
    83,
    84,
    85,
    90,
    91,
    92,
    100,
    101,
    102,
    110,
    111,
    112,
    121,
    130,
    131,
    132,
    140,
    141,
    142,
    143,
    144,
    145,
    150,
    151,
]


class _AuditLogChange_Str(TypedDict):
    key: Literal[
        "name",
        "description",
        "preferred_locale",
        "vanity_url_code",
        "topic",
        "code",
        "allow",
        "deny",
        "permissions",
        "tags",
    ]
    new_value: NotRequired[str]
    old_value: NotRequired[str]


class _AuditLogChange_AssetHash(TypedDict):
    key: Literal[
        "icon_hash",
        "splash_hash",
        "discovery_splash_hash",
        "banner_hash",
        "avatar_hash",
        "asset",
    ]
    new_value: NotRequired[str]
    old_value: NotRequired[str]


class _AuditLogChange_Snowflake(TypedDict):
    key: Literal[
        "id",
        "owner_id",
        "afk_channel_id",
        "rules_channel_id",
        "public_updates_channel_id",
        "widget_channel_id",
        "system_channel_id",
        "application_id",
        "channel_id",
        "inviter_id",
        "guild_id",
    ]
    new_value: NotRequired[Snowflake]
    old_value: NotRequired[Snowflake]


class _AuditLogChange_Bool(TypedDict):
    key: Literal[
        "widget_enabled",
        "nsfw",
        "hoist",
        "mentionable",
        "temporary",
        "deaf",
        "mute",
        "nick",
        "enabled_emoticons",
        "region",
        "rtc_region",
        "available",
        "archived",
        "locked",
        "premium_progress_bar_enabled",
        "enabled",
    ]
    new_value: NotRequired[bool]
    old_value: NotRequired[bool]


class _AuditLogChange_Int(TypedDict):
    key: Literal[
        "afk_timeout",
        "prune_delete_days",
        "position",
        "bitrate",
        "rate_limit_per_user",
        "color",
        "max_uses",
        "max_age",
        "user_limit",
        "auto_archive_duration",
        "default_auto_archive_duration",
    ]
    new_value: NotRequired[int]
    old_value: NotRequired[int]


class _AuditLogChange_ListSnowflake(TypedDict):
    key: Literal["exempt_roles", "exempt_channels"]
    new_value: NotRequired[List[Snowflake]]
    old_value: NotRequired[List[Snowflake]]


class _AuditLogChange_ListRole(TypedDict):
    key: Literal["$add", "$remove"]
    new_value: NotRequired[List[Role]]
    old_value: NotRequired[List[Role]]


class _AuditLogChange_MFALevel(TypedDict):
    key: Literal["mfa_level"]
    new_value: NotRequired[MFALevel]
    old_value: NotRequired[MFALevel]


class _AuditLogChange_VerificationLevel(TypedDict):
    key: Literal["verification_level"]
    new_value: NotRequired[VerificationLevel]
    old_value: NotRequired[VerificationLevel]


class _AuditLogChange_ExplicitContentFilter(TypedDict):
    key: Literal["explicit_content_filter"]
    new_value: NotRequired[ExplicitContentFilterLevel]
    old_value: NotRequired[ExplicitContentFilterLevel]


class _AuditLogChange_DefaultMessageNotificationLevel(TypedDict):
    key: Literal["default_message_notifications"]
    new_value: NotRequired[DefaultMessageNotificationLevel]
    old_value: NotRequired[DefaultMessageNotificationLevel]


class _AuditLogChange_ChannelType(TypedDict):
    key: Literal["type"]
    new_value: NotRequired[ChannelType]
    old_value: NotRequired[ChannelType]


class _AuditLogChange_IntegrationExpireBehaviour(TypedDict):
    key: Literal["expire_behavior"]
    new_value: NotRequired[IntegrationExpireBehavior]
    old_value: NotRequired[IntegrationExpireBehavior]


class _AuditLogChange_VideoQualityMode(TypedDict):
    key: Literal["video_quality_mode"]
    new_value: NotRequired[VideoQualityMode]
    old_value: NotRequired[VideoQualityMode]


class _AuditLogChange_Overwrites(TypedDict):
    key: Literal["permission_overwrites"]
    new_value: NotRequired[List[PermissionOverwrite]]
    old_value: NotRequired[List[PermissionOverwrite]]


class _AuditLogChange_Datetime(TypedDict):
    key: Literal["communication_disabled_until"]
    new_value: NotRequired[datetime.datetime]
    old_value: NotRequired[datetime.datetime]


class _AuditLogChange_ApplicationCommandPermissions(TypedDict):
    key: str
    new_value: NotRequired[ApplicationCommandPermissions]
    old_value: NotRequired[ApplicationCommandPermissions]


class _AuditLogChange_AutoModTriggerType(TypedDict):
    key: Literal["trigger_type"]
    new_value: NotRequired[AutoModTriggerType]
    old_value: NotRequired[AutoModTriggerType]


class _AuditLogChange_AutoModEventType(TypedDict):
    key: Literal["event_type"]
    new_value: NotRequired[AutoModEventType]
    old_value: NotRequired[AutoModEventType]


class _AuditLogChange_AutoModActions(TypedDict):
    key: Literal["actions"]
    new_value: NotRequired[List[AutoModAction]]
    old_value: NotRequired[List[AutoModAction]]


class _AuditLogChange_AutoModTriggerMetadata(TypedDict):
    key: Literal["trigger_metadata"]
    new_value: NotRequired[AutoModTriggerMetadata]
    old_value: NotRequired[AutoModTriggerMetadata]


AuditLogChange = Union[
    _AuditLogChange_Str,
    _AuditLogChange_AssetHash,
    _AuditLogChange_Snowflake,
    _AuditLogChange_Int,
    _AuditLogChange_Bool,
    _AuditLogChange_ListSnowflake,
    _AuditLogChange_ListRole,
    _AuditLogChange_MFALevel,
    _AuditLogChange_VerificationLevel,
    _AuditLogChange_ExplicitContentFilter,
    _AuditLogChange_DefaultMessageNotificationLevel,
    _AuditLogChange_ChannelType,
    _AuditLogChange_IntegrationExpireBehaviour,
    _AuditLogChange_VideoQualityMode,
    _AuditLogChange_Overwrites,
    _AuditLogChange_Datetime,
    _AuditLogChange_ApplicationCommandPermissions,
    _AuditLogChange_AutoModTriggerType,
    _AuditLogChange_AutoModEventType,
    _AuditLogChange_AutoModActions,
    _AuditLogChange_AutoModTriggerMetadata,
]


# All of these are technically only required for matching event types,
# but they're typed as required keys for simplicity
class AuditEntryInfo(TypedDict):
    delete_member_days: str
    members_removed: str
    channel_id: Optional[Snowflake]
    message_id: Snowflake
    count: str
    id: Snowflake
    type: Literal["0", "1"]
    role_name: str
    application_id: Snowflake
    auto_moderation_rule_name: str
    auto_moderation_rule_trigger_type: str
    integration_type: str


class AuditLogEntry(TypedDict):
    target_id: Optional[str]
    changes: NotRequired[List[AuditLogChange]]
    user_id: Optional[Snowflake]
    id: Snowflake
    action_type: AuditLogEvent
    options: NotRequired[AuditEntryInfo]
    reason: NotRequired[str]


class AuditLog(TypedDict):
    audit_log_entries: List[AuditLogEntry]
    application_commands: List[ApplicationCommand]
    auto_moderation_rules: List[AutoModRule]
    guild_scheduled_events: List[GuildScheduledEvent]
    integrations: List[PartialIntegration]
    threads: List[Thread]
    users: List[User]
    webhooks: List[Webhook]



# SPDX-License-Identifier: MIT

from typing import List, Literal, Optional, TypedDict

from typing_extensions import NotRequired

from .snowflake import Snowflake

SubscriptionStatus = Literal[0, 1, 2]


class Subscription(TypedDict):
    id: Snowflake
    user_id: Snowflake
    sku_ids: List[Snowflake]
    entitlement_ids: List[Snowflake]
    renewal_sku_ids: Optional[List[Snowflake]]
    current_period_start: str
    current_period_end: str
    status: SubscriptionStatus
    canceled_at: Optional[str]
    # this is always missing unless queried with a private OAuth scope.
    country: NotRequired[str]



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Literal, Optional, TypedDict

from typing_extensions import NotRequired

from .channel import PartialChannel
from .snowflake import Snowflake
from .user import User


class SourceGuild(TypedDict):
    id: int
    name: str
    icon: str


WebhookType = Literal[1, 2, 3]


class FollowerWebhook(TypedDict):
    channel_id: Snowflake
    webhook_id: Snowflake
    source_channel: NotRequired[PartialChannel]
    source_guild: NotRequired[SourceGuild]


class PartialWebhook(TypedDict):
    id: Snowflake
    type: WebhookType
    guild_id: NotRequired[Optional[Snowflake]]
    user: NotRequired[User]
    token: NotRequired[str]


# note: `total=False` to support `Webhook.partial`
class Webhook(PartialWebhook, total=False):
    name: Optional[str]
    avatar: Optional[str]
    channel_id: Snowflake
    application_id: Optional[Snowflake]



# SPDX-License-Identifier: MIT

import re

from setuptools import setup

version = ""
with open("disnake/__init__.py", encoding="utf-8") as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)  # type: ignore

if not version:
    raise RuntimeError("version is not set")

if version.endswith(("a", "b", "rc")):
    # append version identifier based on commit count
    try:
        import subprocess  # noqa: TID251

        p = subprocess.Popen(
            ["git", "rev-list", "--count", "HEAD"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = p.communicate()
        if out:
            version += out.decode("utf-8").strip()
        p = subprocess.Popen(
            ["git", "rev-parse", "--short", "HEAD"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = p.communicate()
        if out:
            version += "+g" + out.decode("utf-8").strip()
    except Exception:
        pass

packages = [
    "disnake",
    "disnake.bin",
    "disnake.types",
    "disnake.ui",
    "disnake.ui.select",
    "disnake.webhook",
    "disnake.interactions",
    "disnake.ext.commands",
    "disnake.ext.tasks",
    "disnake.ext.mypy_plugin",
]

setup(
    version=version,
    packages=packages,
    include_package_data=True,
)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import functools
import operator
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from .enums import UserFlags
from .utils import MISSING, _generated

if TYPE_CHECKING:
    from typing_extensions import Self


__all__ = (
    "SystemChannelFlags",
    "MessageFlags",
    "PublicUserFlags",
    "Intents",
    "MemberCacheFlags",
    "ApplicationFlags",
    "ChannelFlags",
    "AutoModKeywordPresets",
    "MemberFlags",
    "RoleFlags",
    "AttachmentFlags",
    "SKUFlags",
    "ApplicationInstallTypes",
    "InteractionContextTypes",
)

BF = TypeVar("BF", bound="BaseFlags")
T = TypeVar("T", bound="BaseFlags")


class flag_value(Generic[T]):
    def __init__(self, func: Callable[[Any], int]) -> None:
        self.flag = func(None)
        self.__doc__ = func.__doc__
        self._parent: Type[T] = MISSING

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, flag_value):
            return self.flag == other.flag
        if isinstance(other, BaseFlags):
            return self._parent is other.__class__ and self.flag == other.value
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __or__(self, other: Union[flag_value[T], T]) -> T:
        if isinstance(other, BaseFlags):
            if self._parent is not other.__class__:
                raise TypeError(
                    f"unsupported operand type(s) for |: flags of '{self._parent.__name__}' and flags of '{other.__class__.__name__}'"
                )
            return other._from_value(self.flag | other.value)
        if not isinstance(other, flag_value):
            raise TypeError(
                f"unsupported operand type(s) for |: flags of '{self._parent.__name__}' and {other.__class__}"
            )
        if self._parent is not other._parent:
            raise TypeError(
                f"unsupported operand type(s) for |: flags of '{self._parent.__name__}' and flags of '{other._parent.__name__}'"
            )
        return self._parent._from_value(self.flag | other.flag)

    def __invert__(self: flag_value[T]) -> T:
        return ~self._parent._from_value(self.flag)

    @overload
    def __get__(self, instance: None, owner: Type[BF]) -> flag_value[BF]:
        ...

    @overload
    def __get__(self, instance: BF, owner: Type[BF]) -> bool:
        ...

    def __get__(self, instance: Optional[BF], owner: Type[BF]) -> Any:
        if instance is None:
            return self
        return instance._has_flag(self.flag)

    def __set__(self, instance: BaseFlags, value: bool) -> None:
        instance._set_flag(self.flag, value)

    def __repr__(self) -> str:
        return f"<flag_value flag={self.flag!r}>"


class alias_flag_value(flag_value):
    pass


def all_flags_value(flags: Dict[str, int]) -> int:
    return functools.reduce(operator.or_, flags.values())


class BaseFlags:
    VALID_FLAGS: ClassVar[Dict[str, int]]
    DEFAULT_VALUE: ClassVar[int]

    value: int

    __slots__ = ("value",)

    def __init__(self, **kwargs: bool) -> None:
        self.value = self.DEFAULT_VALUE
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid flag name.")
            setattr(self, key, value)

    @classmethod
    def __init_subclass__(cls, inverted: bool = False, no_fill_flags: bool = False):
        # add a way to bypass filling flags, eg for ListBaseFlags.
        if no_fill_flags:
            return cls

        # use the parent's current flags as a base if they exist
        cls.VALID_FLAGS = getattr(cls, "VALID_FLAGS", {}).copy()

        for name, value in cls.__dict__.items():
            if isinstance(value, flag_value):
                value._parent = cls
                cls.VALID_FLAGS[name] = value.flag

        if not cls.VALID_FLAGS:
            raise RuntimeError(
                "At least one flag must be defined in a BaseFlags subclass, or 'no_fill_flags' must be set to True"
            )

        cls.DEFAULT_VALUE = all_flags_value(cls.VALID_FLAGS) if inverted else 0

        return cls

    @classmethod
    def _from_value(cls, value: int) -> Self:
        self = cls.__new__(cls)
        self.value = value
        return self

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self.value == other.value
        if isinstance(other, flag_value):
            return self.__class__ is other._parent and self.value == other.flag
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __and__(self, other: Self) -> Self:
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for &: '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return self._from_value(self.value & other.value)

    def __iand__(self, other: Self) -> Self:
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for &=: '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        self.value &= other.value
        return self

    def __or__(self, other: Union[Self, flag_value[Self]]) -> Self:
        if isinstance(other, flag_value):
            if self.__class__ is not other._parent:
                raise TypeError(
                    f"unsupported operand type(s) for |: flags of '{self.__class__.__name__}' and flags of '{other._parent.__name__}'"
                )
            return self._from_value(self.value | other.flag)
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for |: '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return self._from_value(self.value | other.value)

    def __ior__(self, other: Union[Self, flag_value[Self]]) -> Self:
        if isinstance(other, flag_value):
            if self.__class__ is not other._parent:
                raise TypeError(
                    f"unsupported operand type(s) for |=: flags of '{self.__class__.__name__}' and flags of '{other._parent.__name__}'"
                )
            self.value |= other.flag
            return self
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for |=: '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        self.value |= other.value
        return self

    def __xor__(self, other: Union[Self, flag_value[Self]]) -> Self:
        if isinstance(other, flag_value):
            if self.__class__ is not other._parent:
                raise TypeError(
                    f"unsupported operand type(s) for ^: flags of '{self.__class__.__name__}' and flags of '{other._parent.__name__}'"
                )
            return self._from_value(self.value ^ other.flag)
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for ^: '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return self._from_value(self.value ^ other.value)

    def __ixor__(self, other: Union[Self, flag_value[Self]]) -> Self:
        if isinstance(other, flag_value):
            if self.__class__ is not other._parent:
                raise TypeError(
                    f"unsupported operand type(s) for ^=: flags of '{self.__class__.__name__}' and flags of '{other._parent.__name__}'"
                )
            self.value ^= other.flag
            return self
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for ^=: '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        self.value ^= other.value
        return self

    def __le__(self, other: Self) -> bool:
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"'<=' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return (self.value & other.value) == self.value

    def __ge__(self, other: Self) -> bool:
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"'>=' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return (self.value | other.value) == self.value

    def __lt__(self, other: Self) -> bool:
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"'<' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return (self.value & other.value) == self.value and self.value != other.value

    def __gt__(self, other: Self) -> bool:
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"'>' not supported between instances of '{self.__class__.__name__}' and '{other.__class__.__name__}'"
            )
        return (self.value | other.value) == self.value and self.value != other.value

    def __invert__(self) -> Self:
        # invert the bit but make sure all truthy values are valid flags
        # this code means that if a flag class doesn't define 1 << 2 that
        # value won't suddenly be set to True
        bitmask = all_flags_value(self.VALID_FLAGS)
        return self._from_value((self.value ^ bitmask) & bitmask)

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} value={self.value}>"

    def __iter__(self) -> Iterator[Tuple[str, bool]]:
        for name, value in self.__class__.__dict__.items():
            if isinstance(value, alias_flag_value):
                continue

            if isinstance(value, flag_value):
                yield (name, self._has_flag(value.flag))

    def _has_flag(self, o: int) -> bool:
        return (self.value & o) == o

    def _set_flag(self, o: int, toggle: bool) -> None:
        if toggle is True:
            self.value |= o
        elif toggle is False:
            self.value &= ~o
        else:
            raise TypeError(f"Value to set for {self.__class__.__name__} must be a bool.")


class ListBaseFlags(BaseFlags, no_fill_flags=True):
    """A base class for flags that aren't powers of 2.
    Instead, values are used as exponents to map to powers of 2 to avoid collisions,
    and only the combined value is stored, which allows all bitwise operations to work as expected.
    """

    __slots__ = ()

    @classmethod
    def _from_values(cls, values: Sequence[int]):
        self = cls.__new__(cls)
        value = 0
        for n in values:
            # protect against DoS with large shift values
            # n.b. performance overhead of this is negligible
            if not (0 <= n < 64):
                raise ValueError("Flag values must be within [0, 64)")
            value += 1 << n
        self.value = value
        return self

    @property
    def values(self) -> List[int]:
        # This essentially converts an int like `0b100110` into `[1, 2, 5]`,
        # i.e. the exponents of set bits in `self.value`.
        # This may look weird but interestingly it's by far the
        # fastest out of all benchmarked snippets, see https://stackoverflow.com/a/49592515/5080607
        # (this code is a derivative of one of them)
        return [i for i, c in enumerate(bin(self.value)[:1:-1]) if c == "1"]

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} values={self.values}>"


class SystemChannelFlags(BaseFlags, inverted=True):
    """Wraps up a Discord system channel flag value.

    Similar to :class:`Permissions`\\, the properties provided are two way.
    You can set and retrieve individual bits using the properties as if they
    were regular bools. This allows you to edit the system flags easily.

    To construct an object you can pass keyword arguments denoting the flags
    to enable or disable.
    Arguments are applied in order, similar to :class:`Permissions`.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two SystemChannelFlags instances are equal.
        .. describe:: x != y

            Checks if two SystemChannelFlags instances are not equal.
        .. describe:: x <= y

            Checks if a SystemChannelFlags instance is a subset of another SystemChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x >= y

            Checks if a SystemChannelFlags instance is a superset of another SystemChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x < y

            Checks if a SystemChannelFlags instance is a strict subset of another SystemChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x > y

            Checks if a SystemChannelFlags instance is a strict superset of another SystemChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x | y, x |= y

            Returns a new SystemChannelFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x & y, x &= y

            Returns a new SystemChannelFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x ^ y, x ^= y

            Returns a new SystemChannelFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: ~x

            Returns a new SystemChannelFlags instance with all flags from x inverted.

            .. versionadded:: 2.6
        .. describe:: hash(x)

               Return the flag's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.


        Additionally supported are a few operations on class attributes.

        .. describe:: SystemChannelFlags.y | SystemChannelFlags.z, SystemChannelFlags(y=True) | SystemChannelFlags.z

            Returns a SystemChannelFlags instance with all provided flags enabled.

        .. describe:: ~SystemChannelFlags.y

            Returns a SystemChannelFlags instance with all flags except ``y`` inverted from their default value.

    Attributes
    ----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self,
            *,
            guild_reminder_notifications: bool = ...,
            join_notification_replies: bool = ...,
            join_notifications: bool = ...,
            premium_subscriptions: bool = ...,
            role_subscription_purchase_notification_replies: bool = ...,
            role_subscription_purchase_notifications: bool = ...,
        ) -> None:
            ...

    # For some reason the flags for system channels are "inverted"
    # ergo, if they're set then it means "suppress" (off in the GUI toggle)
    # Since this is counter-intuitive from an API perspective and annoying
    # these will be inverted automatically

    def _has_flag(self, o: int) -> bool:
        return (self.value & o) != o

    def _set_flag(self, o: int, toggle: bool) -> None:
        if toggle is True:
            self.value &= ~o
        elif toggle is False:
            self.value |= o
        else:
            raise TypeError("Value to set for SystemChannelFlags must be a bool.")

    @flag_value
    def join_notifications(self) -> int:
        """:class:`bool`: Returns ``True`` if the system channel is used for member join notifications."""
        return 1 << 0

    @flag_value
    def premium_subscriptions(self) -> int:
        """:class:`bool`: Returns ``True`` if the system channel is used for "Nitro boosting" notifications."""
        return 1 << 1

    @flag_value
    def guild_reminder_notifications(self) -> int:
        """:class:`bool`: Returns ``True`` if the system channel is used for server setup helpful tips notifications.

        .. versionadded:: 2.0
        """
        return 1 << 2

    @flag_value
    def join_notification_replies(self) -> int:
        """:class:`bool`: Returns ``True`` if the system channel shows sticker reply
        buttons for member join notifications.

        .. versionadded:: 2.3
        """
        return 1 << 3

    @flag_value
    def role_subscription_purchase_notifications(self):
        """:class:`bool`: Returns ``True`` if the system channel shows role
        subscription purchase/renewal notifications.

        .. versionadded:: 2.9
        """
        return 1 << 4

    @flag_value
    def role_subscription_purchase_notification_replies(self):
        """:class:`bool`: Returns ``True`` if the system channel shows sticker reply
        buttons for role subscription purchase/renewal notifications.

        .. versionadded:: 2.9
        """
        return 1 << 5


class MessageFlags(BaseFlags):
    """Wraps up a Discord Message flag value.

    See :class:`SystemChannelFlags`.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two MessageFlags instances are equal.
        .. describe:: x != y

            Checks if two MessageFlags instances are not equal.
        .. describe:: x <= y

            Checks if a MessageFlags instance is a subset of another MessageFlags instance.

            .. versionadded:: 2.6
        .. describe:: x >= y

            Checks if a MessageFlags instance is a superset of another MessageFlags instance.

            .. versionadded:: 2.6
        .. describe:: x < y

            Checks if a MessageFlags instance is a strict subset of another MessageFlags instance.

            .. versionadded:: 2.6
        .. describe:: x > y

            Checks if a MessageFlags instance is a strict superset of another MessageFlags instance.

            .. versionadded:: 2.6
        .. describe:: x | y, x |= y

            Returns a new MessageFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x & y, x &= y

            Returns a new MessageFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x ^ y, x ^= y

            Returns a new MessageFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: ~x

            Returns a new MessageFlags instance with all flags from x inverted.

            .. versionadded:: 2.6
        .. describe:: hash(x)

               Return the flag's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.


        Additionally supported are a few operations on class attributes.

        .. describe:: MessageFlags.y | MessageFlags.z, MessageFlags(y=True) | MessageFlags.z

            Returns a MessageFlags instance with all provided flags enabled.

            .. versionadded:: 2.6
        .. describe:: ~MessageFlags.y

            Returns a MessageFlags instance with all flags except ``y`` inverted from their default value.

            .. versionadded:: 2.6

    .. versionadded:: 1.3

    Attributes
    ----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self,
            *,
            crossposted: bool = ...,
            ephemeral: bool = ...,
            failed_to_mention_roles_in_thread: bool = ...,
            has_snapshot: bool = ...,
            has_thread: bool = ...,
            is_crossposted: bool = ...,
            is_voice_message: bool = ...,
            loading: bool = ...,
            source_message_deleted: bool = ...,
            suppress_embeds: bool = ...,
            suppress_notifications: bool = ...,
            urgent: bool = ...,
        ) -> None:
            ...

    @flag_value
    def crossposted(self):
        """:class:`bool`: Returns ``True`` if the message is the original crossposted message."""
        return 1 << 0

    @flag_value
    def is_crossposted(self):
        """:class:`bool`: Returns ``True`` if the message was crossposted from another channel."""
        return 1 << 1

    @flag_value
    def suppress_embeds(self):
        """:class:`bool`: Returns ``True`` if the message's embeds have been suppressed."""
        return 1 << 2

    @flag_value
    def source_message_deleted(self):
        """:class:`bool`: Returns ``True`` if the source message for this crosspost has been deleted."""
        return 1 << 3

    @flag_value
    def urgent(self):
        """:class:`bool`: Returns ``True`` if the message is an urgent message.

        An urgent message is one sent by Discord Trust and Safety.
        """
        return 1 << 4

    @flag_value
    def has_thread(self):
        """:class:`bool`: Returns ``True`` if the message is associated with a thread.

        .. versionadded:: 2.0
        """
        return 1 << 5

    @flag_value
    def ephemeral(self):
        """:class:`bool`: Returns ``True`` if the message is ephemeral.

        .. versionadded:: 2.0
        """
        return 1 << 6

    @flag_value
    def loading(self):
        """:class:`bool`: Returns ``True`` if the message is a deferred
        interaction response and shows a "thinking" state.

        .. versionadded:: 2.3
        """
        return 1 << 7

    @flag_value
    def failed_to_mention_roles_in_thread(self):
        """:class:`bool`: Returns ``True`` if the message failed to
        mention some roles and add their members to the thread.

        .. versionadded:: 2.4
        """
        return 1 << 8

    @flag_value
    def suppress_notifications(self):
        """:class:`bool`: Returns ``True`` if the message does not
        trigger push and desktop notifications.

        .. versionadded:: 2.9
        """
        return 1 << 12

    @flag_value
    def is_voice_message(self):
        """:class:`bool`: Returns ``True`` if the message is a voice message.

        Messages with this flag will have a single audio attachment, and no other content.

        .. versionadded:: 2.9
        """
        return 1 << 13

    @flag_value
    def has_snapshot(self):
        """:class:`bool`: Returns ``True`` if the message is a forward message.

        Messages with this flag will have only the forward data, and no other content.

        .. versionadded:: 2.10
        """
        return 1 << 14


class PublicUserFlags(BaseFlags):
    """Wraps up the Discord User Public flags.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two PublicUserFlags instances are equal.
        .. describe:: x != y

            Checks if two PublicUserFlags instances are not equal.
        .. describe:: x <= y

            Checks if a PublicUserFlags instance is a subset of another PublicUserFlags instance.

            .. versionadded:: 2.6
        .. describe:: x >= y

            Checks if a PublicUserFlags instance is a superset of another PublicUserFlags instance.

            .. versionadded:: 2.6
        .. describe:: x < y

            Checks if a PublicUserFlags instance is a strict subset of another PublicUserFlags instance.

            .. versionadded:: 2.6
        .. describe:: x > y

            Checks if a PublicUserFlags instance is a strict superset of another PublicUserFlags instance.

            .. versionadded:: 2.6
        .. describe:: x | y, x |= y

            Returns a new PublicUserFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x & y, x &= y

            Returns a new PublicUserFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x ^ y, x ^= y

            Returns a new PublicUserFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: ~x

            Returns a new PublicUserFlags instance with all flags from x inverted.

            .. versionadded:: 2.6
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.


        Additionally supported are a few operations on class attributes.

        .. describe:: PublicUserFlags.y | PublicUserFlags.z, PublicUserFlags(y=True) | PublicUserFlags.z

            Returns a PublicUserFlags instance with all provided flags enabled.

            .. versionadded:: 2.6
        .. describe:: ~PublicUserFlags.y

            Returns a PublicUserFlags instance with all flags except ``y`` inverted from their default value.

            .. versionadded:: 2.6

    .. versionadded:: 1.4

    Attributes
    ----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self,
            *,
            active_developer: bool = ...,
            bug_hunter: bool = ...,
            bug_hunter_level_2: bool = ...,
            discord_certified_moderator: bool = ...,
            early_supporter: bool = ...,
            early_verified_bot_developer: bool = ...,
            http_interactions_bot: bool = ...,
            hypesquad: bool = ...,
            hypesquad_balance: bool = ...,
            hypesquad_bravery: bool = ...,
            hypesquad_brilliance: bool = ...,
            moderator_programs_alumni: bool = ...,
            partner: bool = ...,
            spammer: bool = ...,
            staff: bool = ...,
            system: bool = ...,
            team_user: bool = ...,
            verified_bot: bool = ...,
            verified_bot_developer: bool = ...,
        ) -> None:
            ...

    @flag_value
    def staff(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Employee."""
        return UserFlags.staff.value

    @flag_value
    def partner(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Partner."""
        return UserFlags.partner.value

    @flag_value
    def hypesquad(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Events member."""
        return UserFlags.hypesquad.value

    @flag_value
    def bug_hunter(self):
        """:class:`bool`: Returns ``True`` if the user is a Bug Hunter"""
        return UserFlags.bug_hunter.value

    @flag_value
    def hypesquad_bravery(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Bravery member."""
        return UserFlags.hypesquad_bravery.value

    @flag_value
    def hypesquad_brilliance(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Brilliance member."""
        return UserFlags.hypesquad_brilliance.value

    @flag_value
    def hypesquad_balance(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Balance member."""
        return UserFlags.hypesquad_balance.value

    @flag_value
    def early_supporter(self):
        """:class:`bool`: Returns ``True`` if the user is an Early Supporter."""
        return UserFlags.early_supporter.value

    @flag_value
    def team_user(self):
        """:class:`bool`: Returns ``True`` if the user is a Team User."""
        return UserFlags.team_user.value

    @flag_value
    def system(self):
        """:class:`bool`: Returns ``True`` if the user is a system user (i.e. represents Discord officially)."""
        return UserFlags.system.value

    @flag_value
    def bug_hunter_level_2(self):
        """:class:`bool`: Returns ``True`` if the user is a Bug Hunter Level 2"""
        return UserFlags.bug_hunter_level_2.value

    @flag_value
    def verified_bot(self):
        """:class:`bool`: Returns ``True`` if the user is a Verified Bot."""
        return UserFlags.verified_bot.value

    @flag_value
    def verified_bot_developer(self):
        """:class:`bool`: Returns ``True`` if the user is an Early Verified Bot Developer."""
        return UserFlags.verified_bot_developer.value

    @alias_flag_value
    def early_verified_bot_developer(self):
        """:class:`bool`: An alias for :attr:`verified_bot_developer`.

        .. versionadded:: 1.5
        """
        return UserFlags.verified_bot_developer.value

    @flag_value
    def moderator_programs_alumni(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Moderator Programs Alumni.

        .. versionadded:: 2.8
        """
        return UserFlags.discord_certified_moderator.value

    @alias_flag_value
    def discord_certified_moderator(self):
        """:class:`bool`: An alias for :attr:`moderator_programs_alumni`.

        .. versionadded:: 2.0
        """
        return UserFlags.discord_certified_moderator.value

    @flag_value
    def http_interactions_bot(self):
        """:class:`bool`: Returns ``True`` if the user is a bot that only uses HTTP interactions.

        .. versionadded:: 2.3
        """
        return UserFlags.http_interactions_bot.value

    @flag_value
    def spammer(self):
        """:class:`bool`: Returns ``True`` if the user is marked as a spammer.

        .. versionadded:: 2.3
        """
        return UserFlags.spammer.value

    @flag_value
    def active_developer(self):
        """:class:`bool`: Returns ``True`` if the user is an Active Developer.

        .. versionadded:: 2.8
        """
        return UserFlags.active_developer.value

    def all(self) -> List[UserFlags]:
        """List[:class:`UserFlags`]: Returns all public flags the user has."""
        return [public_flag for public_flag in UserFlags if self._has_flag(public_flag.value)]


class Intents(BaseFlags):
    """Wraps up a Discord gateway intent flag.

    Similar to :class:`Permissions`\\, the properties provided are two way.
    You can set and retrieve individual bits using the properties as if they
    were regular bools.

    To construct an object you can pass keyword arguments denoting the flags
    to enable or disable.
    Arguments are applied in order, similar to :class:`Permissions`.

    This is used to disable certain gateway features that are unnecessary to
    run your bot. To make use of this, it is passed to the ``intents`` keyword
    argument of :class:`Client`.

    .. versionadded:: 1.5

    .. collapse:: operations

        .. describe:: x == y

            Checks if two Intents instances are equal.
        .. describe:: x != y

            Checks if two Intents instances are not equal.
        .. describe:: x <= y

            Checks if an Intents instance is a subset of another Intents instance.

            .. versionadded:: 2.6
        .. describe:: x >= y

            Checks if an Intents instance is a superset of another Intents instance.

            .. versionadded:: 2.6
        .. describe:: x < y

            Checks if an Intents instance is a strict subset of another Intents instance.

            .. versionadded:: 2.6
        .. describe:: x > y

            Checks if an Intents instance is a strict superset of another Intents instance.

            .. versionadded:: 2.6
        .. describe:: x | y, x |= y

            Returns a new Intents instance with all enabled intents from both x and y.
            (Using ``|=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x & y, x &= y

            Returns a new Intents instance with only intents enabled on both x and y.
            (Using ``&=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x ^ y, x ^= y

            Returns a new Intents instance with only intents enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: ~x

            Returns a new Intents instance with all intents inverted from x.

            .. versionadded:: 2.6
        .. describe:: hash(x)

               Return the flag's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.


        Additionally supported are a few operations on class attributes.

        .. describe:: Intents.y | Intents.z, Intents(y=True) | Intents.z

            Returns an Intents instance with all provided flags enabled.

            .. versionadded:: 2.6
        .. describe:: ~Intents.y

            Returns an Intents instance with all flags except ``y`` inverted from their default value.

            .. versionadded:: 2.6

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.

        .. versionchanged:: 2.6

            This can be now be provided on initialisation.
    """

    __slots__ = ()

    @overload
    @_generated
    def __init__(
        self,
        value: Optional[int] = None,
        *,
        automod: bool = ...,
        automod_configuration: bool = ...,
        automod_execution: bool = ...,
        bans: bool = ...,
        dm_messages: bool = ...,
        dm_polls: bool = ...,
        dm_reactions: bool = ...,
        dm_typing: bool = ...,
        emojis: bool = ...,
        emojis_and_stickers: bool = ...,
        expressions: bool = ...,
        guild_messages: bool = ...,
        guild_polls: bool = ...,
        guild_reactions: bool = ...,
        guild_scheduled_events: bool = ...,
        guild_typing: bool = ...,
        guilds: bool = ...,
        integrations: bool = ...,
        invites: bool = ...,
        members: bool = ...,
        message_content: bool = ...,
        messages: bool = ...,
        moderation: bool = ...,
        polls: bool = ...,
        presences: bool = ...,
        reactions: bool = ...,
        typing: bool = ...,
        voice_states: bool = ...,
        webhooks: bool = ...,
    ) -> None:
        ...

    @overload
    @_generated
    def __init__(self: NoReturn) -> None:
        ...

    def __init__(self, value: Optional[int] = None, **kwargs: bool) -> None:
        if value is not None:
            if not isinstance(value, int):
                raise TypeError(
                    f"Expected int, received {type(value).__name__} for argument 'value'."
                )
            if value < 0:
                raise ValueError("Expected a non-negative value.")
            self.value = value
        else:
            self.value = self.DEFAULT_VALUE
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid flag name.")
            setattr(self, key, value)

    @classmethod
    def all(cls) -> Self:
        """A factory method that creates an :class:`Intents` instance with everything enabled."""
        self = cls.__new__(cls)
        self.value = all_flags_value(cls.VALID_FLAGS)
        return self

    @classmethod
    def none(cls) -> Self:
        """A factory method that creates an :class:`Intents` instance with everything disabled."""
        self = cls.__new__(cls)
        self.value = self.DEFAULT_VALUE
        return self

    @classmethod
    def default(cls) -> Self:
        """A factory method that creates an :class:`Intents` instance with everything enabled
        except :attr:`presences`, :attr:`members`, and :attr:`message_content`.
        """
        self = cls.all()
        self.presences = False
        self.members = False
        self.message_content = False
        return self

    @flag_value
    def guilds(self):
        """:class:`bool`: Whether guild related events are enabled.

        This corresponds to the following events:

        - :func:`on_guild_join`
        - :func:`on_guild_remove`
        - :func:`on_guild_available`
        - :func:`on_guild_unavailable`
        - :func:`on_guild_channel_update`
        - :func:`on_guild_channel_create`
        - :func:`on_guild_channel_delete`
        - :func:`on_guild_channel_pins_update`

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Client.guilds`
        - :class:`Guild` and all its attributes.
        - :meth:`Client.get_channel`
        - :meth:`Client.get_all_channels`

        It is highly advisable to leave this intent enabled for your bot to function.
        """
        return 1 << 0

    @flag_value
    def members(self):
        """:class:`bool`: Whether guild member related events are enabled.

        This corresponds to the following events:

        - :func:`on_member_join`
        - :func:`on_member_remove`
        - :func:`on_member_update`
        - :func:`on_user_update`
        - :func:`on_guild_scheduled_event_subscribe`
        - :func:`on_guild_scheduled_event_unsubscribe`

        This also corresponds to the following attributes and classes in terms of cache:

        - :meth:`Client.get_all_members`
        - :meth:`Client.get_user`
        - :meth:`Guild.chunk`
        - :meth:`Guild.fetch_members`
        - :meth:`Guild.get_member`
        - :attr:`Guild.members`
        - :attr:`Member.roles`
        - :attr:`Member.nick`
        - :attr:`Member.premium_since`
        - :attr:`User.name`
        - :attr:`User.avatar`
        - :attr:`User.discriminator`
        - :attr:`User.global_name`

        For more information go to the :ref:`member intent documentation <need_members_intent>`.

        .. note::

            Currently, this requires opting in explicitly via the developer portal as well.
            Bots in over 100 guilds will need to apply to Discord for verification.
        """
        return 1 << 1

    @flag_value
    def moderation(self):
        """:class:`bool`: Whether guild moderation related events are enabled.

        This corresponds to the following events:

        - :func:`on_member_ban`
        - :func:`on_member_unban`
        - :func:`on_audit_log_entry_create`

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 2

    @alias_flag_value
    def bans(self):
        """:class:`bool`: Alias of :attr:`.moderation`.

        .. versionchanged:: 2.8
            Changed to an alias.
        """
        return 1 << 2

    @flag_value
    def expressions(self):
        """:class:`bool`: Whether events related to guild emojis, stickers, and
        soundboard sounds are enabled.

        .. versionadded:: 2.10

        This corresponds to the following events:

        - :func:`on_guild_emojis_update`
        - :func:`on_guild_stickers_update`
        - :func:`on_guild_soundboard_sounds_update`

        This also corresponds to the following attributes and classes in terms of cache:

        - :class:`Emoji`
        - :class:`GuildSticker`
        - :class:`GuildSoundboardSound`
        - :meth:`Client.get_emoji`
        - :meth:`Client.get_sticker`
        - :meth:`Client.get_soundboard_sound`
        - :meth:`Client.emojis`
        - :meth:`Client.stickers`
        - :meth:`Client.soundboard_sounds`
        - :attr:`Guild.emojis`
        - :attr:`Guild.stickers`
        - :attr:`Guild.soundboard_sounds`
        """
        return 1 << 3

    @alias_flag_value
    def emojis_and_stickers(self):
        """:class:`bool`: Alias of :attr:`.expressions`.

        .. versionadded:: 2.0

        .. versionchanged:: 2.10
            Changed to an alias.
        """
        return 1 << 3

    @alias_flag_value
    def emojis(self):
        """:class:`bool`: Alias of :attr:`.expressions`.

        .. versionchanged:: 2.0
            Changed to an alias.
        """
        return 1 << 3

    @flag_value
    def integrations(self):
        """:class:`bool`: Whether guild integration related events are enabled.

        This corresponds to the following events:

        - :func:`on_guild_integrations_update`
        - :func:`on_integration_create`
        - :func:`on_integration_update`
        - :func:`on_raw_integration_delete`

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 4

    @flag_value
    def webhooks(self):
        """:class:`bool`: Whether guild webhook related events are enabled.

        This corresponds to the following events:

        - :func:`on_webhooks_update`

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 5

    @flag_value
    def invites(self):
        """:class:`bool`: Whether guild invite related events are enabled.

        This corresponds to the following events:

        - :func:`on_invite_create`
        - :func:`on_invite_delete`

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 6

    @flag_value
    def voice_states(self):
        """:class:`bool`: Whether guild voice state related events are enabled.

        This corresponds to the following events:

        - :func:`on_voice_state_update`
        - :func:`on_voice_channel_effect`
        - :func:`on_raw_voice_channel_effect`

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`VoiceChannel.members`
        - :attr:`VoiceChannel.voice_states`
        - :attr:`Member.voice`

        .. note::

            This intent is required to connect to voice.
        """
        return 1 << 7

    @flag_value
    def presences(self):
        """:class:`bool`: Whether guild presence related events are enabled.

        This corresponds to the following events:

        - :func:`on_presence_update`
        - :func:`on_raw_presence_update`

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Member.activities`
        - :attr:`Member.status`
        - :attr:`Member.raw_status`

        For more information go to the :ref:`presence intent documentation <need_presence_intent>`.

        .. note::

            Currently, this requires opting in explicitly via the developer portal as well.
            Bots in over 100 guilds will need to apply to Discord for verification.
        """
        return 1 << 8

    @alias_flag_value
    def messages(self):
        """:class:`bool`: Whether guild and direct message related events are enabled.

        This is a shortcut to set or get both :attr:`guild_messages` and :attr:`dm_messages`.

        This corresponds to the following events:

        - :func:`on_message` (both guilds and DMs)
        - :func:`on_message_edit` (both guilds and DMs)
        - :func:`on_message_delete` (both guilds and DMs)
        - :func:`on_raw_message_delete` (both guilds and DMs)
        - :func:`on_raw_message_edit` (both guilds and DMs)

        This also corresponds to the following attributes and classes in terms of cache:

        - :class:`Message`
        - :attr:`Client.cached_messages`

        Note that due to an implicit relationship this also corresponds to the following events:

        - :func:`on_reaction_add` (both guilds and DMs)
        - :func:`on_reaction_remove` (both guilds and DMs)
        - :func:`on_reaction_clear` (both guilds and DMs)

        .. note::

            :attr:`.Intents.message_content` is required to receive the content of messages.
        """
        return (1 << 9) | (1 << 12)

    @flag_value
    def guild_messages(self):
        """:class:`bool`: Whether guild message related events are enabled.

        See also :attr:`dm_messages` for DMs or :attr:`messages` for both.

        This corresponds to the following events:

        - :func:`on_message` (only for guilds)
        - :func:`on_message_edit` (only for guilds)
        - :func:`on_message_delete` (only for guilds)
        - :func:`on_raw_message_delete` (only for guilds)
        - :func:`on_raw_message_edit` (only for guilds)

        This also corresponds to the following attributes and classes in terms of cache:

        - :class:`Message`
        - :attr:`Client.cached_messages` (only for guilds)

        Note that due to an implicit relationship this also corresponds to the following events:

        - :func:`on_reaction_add` (only for guilds)
        - :func:`on_reaction_remove` (only for guilds)
        - :func:`on_reaction_clear` (only for guilds)
        """
        return 1 << 9

    @flag_value
    def dm_messages(self):
        """:class:`bool`: Whether direct message related events are enabled.

        See also :attr:`guild_messages` for guilds or :attr:`messages` for both.

        This corresponds to the following events:

        - :func:`on_message` (only for DMs)
        - :func:`on_message_edit` (only for DMs)
        - :func:`on_message_delete` (only for DMs)
        - :func:`on_raw_message_delete` (only for DMs)
        - :func:`on_raw_message_edit` (only for DMs)

        This also corresponds to the following attributes and classes in terms of cache:

        - :class:`Message`
        - :attr:`Client.cached_messages` (only for DMs)

        Note that due to an implicit relationship this also corresponds to the following events:

        - :func:`on_reaction_add` (only for DMs)
        - :func:`on_reaction_remove` (only for DMs)
        - :func:`on_reaction_clear` (only for DMs)
        """
        return 1 << 12

    @flag_value
    def message_content(self):
        """:class:`bool`: Whether messages will have access to message content.

        .. versionadded:: 2.5

        This applies to the following fields on :class:`~disnake.Message` instances:

        - :attr:`~disnake.Message.content`
        - :attr:`~disnake.Message.embeds`
        - :attr:`~disnake.Message.attachments`
        - :attr:`~disnake.Message.components`

        The following cases will always have the above fields:

        - Messages the bot sends
        - Messages the bot receives as a direct message
        - Messages in which the bot is mentioned
        - Messages received from an interaction payload, these will typically be attributes on :class:`~disnake.MessageInteraction` instances.

        In addition, this also corresponds to the following fields:

        - :attr:`AutoModActionExecution.content`
        - :attr:`AutoModActionExecution.matched_content`

        For more information go to the :ref:`message content intent documentation <need_message_content_intent>`.

        .. note::

            Currently, this requires opting in explicitly via the developer portal as well.
            Bots in over 100 guilds will need to apply to Discord for verification.
        """
        return 1 << 15

    @alias_flag_value
    def reactions(self):
        """:class:`bool`: Whether guild and direct message reaction related events are enabled.

        This is a shortcut to set or get both :attr:`guild_reactions` and :attr:`dm_reactions`.

        This corresponds to the following events:

        - :func:`on_reaction_add` (both guilds and DMs)
        - :func:`on_reaction_remove` (both guilds and DMs)
        - :func:`on_reaction_clear` (both guilds and DMs)
        - :func:`on_raw_reaction_add` (both guilds and DMs)
        - :func:`on_raw_reaction_remove` (both guilds and DMs)
        - :func:`on_raw_reaction_clear` (both guilds and DMs)

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Message.reactions` (both guild and DM messages)
        """
        return (1 << 10) | (1 << 13)

    @flag_value
    def guild_reactions(self):
        """:class:`bool`: Whether guild reaction related events are enabled.

        See also :attr:`dm_reactions` for DMs or :attr:`reactions` for both.

        This corresponds to the following events:

        - :func:`on_reaction_add` (only for guilds)
        - :func:`on_reaction_remove` (only for guilds)
        - :func:`on_reaction_clear` (only for guilds)
        - :func:`on_raw_reaction_add` (only for guilds)
        - :func:`on_raw_reaction_remove` (only for guilds)
        - :func:`on_raw_reaction_clear` (only for guilds)

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Message.reactions` (only for guild messages)
        """
        return 1 << 10

    @flag_value
    def dm_reactions(self):
        """:class:`bool`: Whether direct message reaction related events are enabled.

        See also :attr:`guild_reactions` for guilds or :attr:`reactions` for both.

        This corresponds to the following events:

        - :func:`on_reaction_add` (only for DMs)
        - :func:`on_reaction_remove` (only for DMs)
        - :func:`on_reaction_clear` (only for DMs)
        - :func:`on_raw_reaction_add` (only for DMs)
        - :func:`on_raw_reaction_remove` (only for DMs)
        - :func:`on_raw_reaction_clear` (only for DMs)

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Message.reactions` (only for DM messages)
        """
        return 1 << 13

    @alias_flag_value
    def typing(self):
        """:class:`bool`: Whether guild and direct message typing related events are enabled.

        This is a shortcut to set or get both :attr:`guild_typing` and :attr:`dm_typing`.

        This corresponds to the following events:

        - :func:`on_typing` (both guilds and DMs)

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return (1 << 11) | (1 << 14)

    @flag_value
    def guild_typing(self):
        """:class:`bool`: Whether guild typing related events are enabled.

        See also :attr:`dm_typing` for DMs or :attr:`typing` for both.

        This corresponds to the following events:

        - :func:`on_typing` (only for guilds)

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 11

    @flag_value
    def dm_typing(self):
        """:class:`bool`: Whether direct message typing related events are enabled.

        See also :attr:`guild_typing` for guilds or :attr:`typing` for both.

        This corresponds to the following events:

        - :func:`on_typing` (only for DMs)

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 14

    @flag_value
    def guild_scheduled_events(self):
        """:class:`bool`: Whether guild scheduled event related events are enabled.

        .. versionadded:: 2.3

        This corresponds to the following events:

        - :func:`on_guild_scheduled_event_create`
        - :func:`on_guild_scheduled_event_delete`
        - :func:`on_guild_scheduled_event_update`
        - :func:`on_guild_scheduled_event_subscribe`
        - :func:`on_guild_scheduled_event_unsubscribe`
        - :func:`on_raw_guild_scheduled_event_subscribe`
        - :func:`on_raw_guild_scheduled_event_unsubscribe`

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Guild.scheduled_events`
        - :meth:`Guild.get_scheduled_event`
        - :attr:`StageInstance.guild_scheduled_event`
        """
        return 1 << 16

    @flag_value
    def automod_configuration(self):
        """:class:`bool`: Whether auto moderation configuration related events are enabled.

        .. versionadded:: 2.6

        This corresponds to the following events:

        - :func:`on_automod_rule_create`
        - :func:`on_automod_rule_delete`
        - :func:`on_automod_rule_update`

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 20

    @flag_value
    def automod_execution(self):
        """:class:`bool`: Whether auto moderation execution related events are enabled.

        .. versionadded:: 2.6

        This corresponds to the following events:

        - :func:`on_automod_action_execution`

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return 1 << 21

    @alias_flag_value
    def automod(self):
        """:class:`bool`: Whether auto moderation related events are enabled.

        .. versionadded:: 2.6

        This is a shortcut to set or get both :attr:`automod_configuration` and :attr:`automod_execution`.

        This corresponds to the following events:

        - :func:`on_automod_rule_create`
        - :func:`on_automod_rule_delete`
        - :func:`on_automod_rule_update`
        - :func:`on_automod_action_execution`

        This does not correspond to any attributes or classes in the library in terms of cache.
        """
        return (1 << 20) | (1 << 21)

    @alias_flag_value
    def polls(self):
        """:class:`bool`: Whether guild and direct message polls related events are enabled.

        This is a shortcut to set or get both :attr:`guild_polls` and :attr:`dm_polls`.

        This corresponds to the following events:

        - :func:`on_poll_vote_add` (both guilds and DMs)
        - :func:`on_poll_vote_remove` (both guilds and DMs)
        - :func:`on_raw_poll_vote_add` (both guilds and DMs)
        - :func:`on_raw_poll_vote_remove` (both guilds and DMs)
        """
        return (1 << 24) | (1 << 25)

    @flag_value
    def guild_polls(self):
        """:class:`bool`: Whether guild polls related events are enabled.

        .. versionadded:: 2.10

        This corresponds to the following events:

        - :func:`on_poll_vote_add` (only for guilds)
        - :func:`on_poll_vote_remove` (only for guilds)
        - :func:`on_raw_poll_vote_add` (only for guilds)
        - :func:`on_raw_poll_vote_remove` (only for guilds)

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Message.poll` (only for guild messages)
        - :class:`Poll` and all its attributes.
        """
        return 1 << 24

    @flag_value
    def dm_polls(self):
        """:class:`bool`: Whether direct message polls related events are enabled.

        .. versionadded:: 2.10

        This corresponds to the following events:

        - :func:`on_poll_vote_add` (only for DMs)
        - :func:`on_poll_vote_remove` (only for DMs)
        - :func:`on_raw_poll_vote_add` (only for DMs)
        - :func:`on_raw_poll_vote_remove` (only for DMs)

        This also corresponds to the following attributes and classes in terms of cache:

        - :attr:`Message.poll` (only for DM messages)
        - :class:`Poll` and all its attributes.
        """
        return 1 << 25


class MemberCacheFlags(BaseFlags):
    """Controls the library's cache policy when it comes to members.

    This allows for finer grained control over what members are cached.
    Note that the bot's own member is always cached. This class is passed
    to the ``member_cache_flags`` parameter in :class:`Client`.

    Due to a quirk in how Discord works, in order to ensure proper cleanup
    of cache resources it is recommended to have :attr:`Intents.members`
    enabled. Otherwise the library cannot know when a member leaves a guild and
    is thus unable to cleanup after itself.

    To construct an object you can pass keyword arguments denoting the flags
    to enable or disable.
    Arguments are applied in order, similar to :class:`Permissions`.

    The default value is all flags enabled.

    .. versionadded:: 1.5

    .. collapse:: operations

        .. describe:: x == y

            Checks if two MemberCacheFlags instances are equal.
        .. describe:: x != y

            Checks if two MemberCacheFlags instances are not equal.
        .. describe:: x <= y

            Checks if a MemberCacheFlags instance is a subset of another MemberCacheFlags instance.

            .. versionadded:: 2.6
        .. describe:: x >= y

            Checks if a MemberCacheFlags instance is a superset of another MemberCacheFlags instance.

            .. versionadded:: 2.6
        .. describe:: x < y

            Checks if a MemberCacheFlags instance is a strict subset of another MemberCacheFlags instance.

            .. versionadded:: 2.6
        .. describe:: x > y

            Checks if a MemberCacheFlags instance is a strict superset of another MemberCacheFlags instance.

            .. versionadded:: 2.6
        .. describe:: x | y, x |= y

            Returns a new MemberCacheFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x & y, x &= y

            Returns a new MemberCacheFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x ^ y, x ^= y

            Returns a new MemberCacheFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: ~x

            Returns a new MemberCacheFlags instance with all flags from x inverted.

            .. versionadded:: 2.6
        .. describe:: hash(x)

               Return the flag's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.


        Additionally supported are a few operations on class attributes.

        .. describe:: MemberCacheFlags.y | MemberCacheFlags.z, MemberCacheFlags(y=True) | MemberCacheFlags.z

            Returns a MemberCacheFlags instance with all provided flags enabled.

            .. versionadded:: 2.6
        .. describe:: ~MemberCacheFlags.y

            Returns a MemberCacheFlags instance with all flags except ``y`` inverted from their default value.

            .. versionadded:: 2.6

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    @overload
    @_generated
    def __init__(self, *, joined: bool = ..., voice: bool = ...) -> None:
        ...

    @overload
    @_generated
    def __init__(self: NoReturn) -> None:
        ...

    def __init__(self, **kwargs: bool) -> None:
        self.value = all_flags_value(self.VALID_FLAGS)
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid flag name.")
            setattr(self, key, value)

    @classmethod
    def all(cls) -> Self:
        """A factory method that creates a :class:`MemberCacheFlags` instance with everything enabled."""
        self = cls.__new__(cls)
        self.value = all_flags_value(cls.VALID_FLAGS)
        return self

    @classmethod
    def none(cls) -> Self:
        """A factory method that creates a :class:`MemberCacheFlags` instance with everything disabled."""
        self = cls.__new__(cls)
        self.value = self.DEFAULT_VALUE
        return self

    @property
    def _empty(self):
        return self.value == self.DEFAULT_VALUE

    @flag_value
    def voice(self) -> int:
        """:class:`bool`: Whether to cache members that are in voice.

        This requires :attr:`Intents.voice_states`.

        Members that leave voice are no longer cached.
        """
        return 1

    @flag_value
    def joined(self) -> int:
        """:class:`bool`: Whether to cache members that joined the guild
        or are chunked as part of the initial log in flow.

        This requires :attr:`Intents.members`.

        Members that leave the guild are no longer cached.
        """
        return 2

    @classmethod
    def from_intents(cls, intents: Intents) -> Self:
        """A factory method that creates a :class:`MemberCacheFlags` instance based on
        the currently selected :class:`Intents`.

        Parameters
        ----------
        intents: :class:`Intents`
            The intents to select from.

        Returns
        -------
        :class:`MemberCacheFlags`
            The resulting member cache flags.
        """
        self = cls.none()
        if intents.members:
            self.joined = True
        if intents.voice_states:
            self.voice = True

        return self

    def _verify_intents(self, intents: Intents) -> None:
        if self.voice and not intents.voice_states:
            raise ValueError("MemberCacheFlags.voice requires Intents.voice_states")

        if self.joined and not intents.members:
            raise ValueError("MemberCacheFlags.joined requires Intents.members")

    @property
    def _voice_only(self):
        return self.value == 1


class ApplicationFlags(BaseFlags):
    """Wraps up the Discord Application flags.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two ApplicationFlags instances are equal.
        .. describe:: x != y

            Checks if two ApplicationFlags instances are not equal.
        .. describe:: x <= y

            Checks if an ApplicationFlags instance is a subset of another ApplicationFlags instance.

            .. versionadded:: 2.6
        .. describe:: x >= y

            Checks if an ApplicationFlags instance is a superset of another ApplicationFlags instance.

            .. versionadded:: 2.6
        .. describe:: x < y

            Checks if an ApplicationFlags instance is a strict subset of another ApplicationFlags instance.

            .. versionadded:: 2.6
        .. describe:: x > y

            Checks if an ApplicationFlags instance is a strict superset of another ApplicationFlags instance.

            .. versionadded:: 2.6
        .. describe:: x | y, x |= y

            Returns a new ApplicationFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x & y, x &= y

            Returns a new ApplicationFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x ^ y, x ^= y

            Returns a new ApplicationFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: ~x

            Returns a new ApplicationFlags instance with all flags from x inverted.

            .. versionadded:: 2.6
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.


        Additionally supported are a few operations on class attributes.

        .. describe:: ApplicationFlags.y | ApplicationFlags.z, ApplicationFlags(y=True) | ApplicationFlags.z

            Returns an ApplicationFlags instance with all provided flags enabled.

            .. versionadded:: 2.6
        .. describe:: ~ApplicationFlags.y

            Returns an ApplicationFlags instance with all flags except ``y`` inverted from their default value.

            .. versionadded:: 2.6

    .. versionadded:: 2.0

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self,
            *,
            application_auto_moderation_rule_create_badge: bool = ...,
            application_command_badge: bool = ...,
            embedded: bool = ...,
            gateway_guild_members: bool = ...,
            gateway_guild_members_limited: bool = ...,
            gateway_message_content: bool = ...,
            gateway_message_content_limited: bool = ...,
            gateway_presence: bool = ...,
            gateway_presence_limited: bool = ...,
            verification_pending_guild_limit: bool = ...,
        ) -> None:
            ...

    @flag_value
    def application_auto_moderation_rule_create_badge(self):
        """:class:`bool`: Returns ``True`` if the application uses the Auto Moderation API."""
        return 1 << 6

    @flag_value
    def gateway_presence(self):
        """:class:`bool`: Returns ``True`` if the application is verified and is allowed to
        receive presence information over the gateway.
        """
        return 1 << 12

    @flag_value
    def gateway_presence_limited(self):
        """:class:`bool`: Returns ``True`` if the application is allowed to receive limited
        presence information over the gateway.
        """
        return 1 << 13

    @flag_value
    def gateway_guild_members(self):
        """:class:`bool`: Returns ``True`` if the application is verified and is allowed to
        receive guild members information over the gateway.
        """
        return 1 << 14

    @flag_value
    def gateway_guild_members_limited(self):
        """:class:`bool`: Returns ``True`` if the application is allowed to receive limited
        guild members information over the gateway.
        """
        return 1 << 15

    @flag_value
    def verification_pending_guild_limit(self):
        """:class:`bool`: Returns ``True`` if the application is currently pending verification
        and has hit the guild limit.
        """
        return 1 << 16

    @flag_value
    def embedded(self):
        """:class:`bool`: Returns ``True`` if the application is embedded within the Discord client."""
        return 1 << 17

    @flag_value
    def gateway_message_content(self):
        """:class:`bool`: Returns ``True`` if the application is verified and is allowed to
        receive message content over the gateway.
        """
        return 1 << 18

    @flag_value
    def gateway_message_content_limited(self):
        """:class:`bool`: Returns ``True`` if the application is verified and is allowed to
        receive limited message content over the gateway.
        """
        return 1 << 19

    @flag_value
    def application_command_badge(self):
        """:class:`bool`: Returns ``True`` if the application has registered global application commands.

        .. versionadded:: 2.6
        """
        return 1 << 23


class ChannelFlags(BaseFlags):
    """Wraps up the Discord Channel flags.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two ChannelFlags instances are equal.
        .. describe:: x != y

            Checks if two ChannelFlags instances are not equal.
        .. describe:: x <= y

            Checks if a ChannelFlags instance is a subset of another ChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x >= y

            Checks if a ChannelFlags instance is a superset of another ChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x < y

            Checks if a ChannelFlags instance is a strict subset of another ChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x > y

            Checks if a ChannelFlags instance is a strict superset of another ChannelFlags instance.

            .. versionadded:: 2.6
        .. describe:: x | y, x |= y

            Returns a new ChannelFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x & y, x &= y

            Returns a new ChannelFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: x ^ y, x ^= y

            Returns a new ChannelFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).

            .. versionadded:: 2.6
        .. describe:: ~x

            Returns a new ChannelFlags instance with all flags from x inverted.

            .. versionadded:: 2.6
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.


        Additionally supported are a few operations on class attributes.

        .. describe:: ChannelFlags.y | ChannelFlags.z, ChannelFlags(y=True) | ChannelFlags.z

            Returns a ChannelFlags instance with all provided flags enabled.

            .. versionadded:: 2.6
        .. describe:: ~ChannelFlags.y

            Returns a ChannelFlags instance with all flags except ``y`` inverted from their default value.

            .. versionadded:: 2.6

    .. versionadded:: 2.5

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self,
            *,
            hide_media_download_options: bool = ...,
            pinned: bool = ...,
            require_tag: bool = ...,
        ) -> None:
            ...

    @flag_value
    def pinned(self):
        """:class:`bool`: Returns ``True`` if the thread is pinned.

        This only applies to threads that are part of a :class:`ForumChannel` or :class:`MediaChannel`.
        """
        return 1 << 1

    @flag_value
    def require_tag(self):
        """:class:`bool`: Returns ``True`` if the channel requires all newly created threads to have a tag.

        This only applies to channels of types :class:`ForumChannel` or :class:`MediaChannel`.

        .. versionadded:: 2.6
        """
        return 1 << 4

    @flag_value
    def hide_media_download_options(self):
        """:class:`bool`: Returns ``True`` if the channel hides the embedded media download options.

        This only applies to channels of type :class:`MediaChannel`.

        .. versionadded:: 2.10
        """
        return 1 << 15


class AutoModKeywordPresets(ListBaseFlags):
    """Wraps up the pre-defined auto moderation keyword lists, provided by Discord.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two AutoModKeywordPresets instances are equal.
        .. describe:: x != y

            Checks if two AutoModKeywordPresets instances are not equal.
        .. describe:: x <= y

            Checks if an AutoModKeywordPresets instance is a subset of another AutoModKeywordPresets instance.
        .. describe:: x >= y

            Checks if an AutoModKeywordPresets instance is a superset of another AutoModKeywordPresets instance.
        .. describe:: x < y

            Checks if an AutoModKeywordPresets instance is a strict subset of another AutoModKeywordPresets instance.
        .. describe:: x > y

            Checks if an AutoModKeywordPresets instance is a strict superset of another AutoModKeywordPresets instance.
        .. describe:: x | y, x |= y

            Returns a new AutoModKeywordPresets instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).
        .. describe:: x & y, x &= y

            Returns a new AutoModKeywordPresets instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).
        .. describe:: x ^ y, x ^= y

            Returns a new AutoModKeywordPresets instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).
        .. describe:: ~x

            Returns a new AutoModKeywordPresets instance with all flags from x inverted.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.


        Additionally supported are a few operations on class attributes.

        .. describe:: AutoModKeywordPresets.y | AutoModKeywordPresets.z, AutoModKeywordPresets(y=True) | AutoModKeywordPresets.z

            Returns an AutoModKeywordPresets instance with all provided flags enabled.

        .. describe:: ~AutoModKeywordPresets.y

            Returns an AutoModKeywordPresets instance with all flags except ``y`` inverted from their default value.

    .. versionadded:: 2.6

    Attributes
    ----------
    values: :class:`int`
        The raw values. You should query flags via the properties
        rather than using these raw values.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self, *, profanity: bool = ..., sexual_content: bool = ..., slurs: bool = ...
        ) -> None:
            ...

    @classmethod
    def all(cls) -> Self:
        """A factory method that creates an :class:`AutoModKeywordPresets` instance with everything enabled."""
        self = cls.__new__(cls)
        self.value = all_flags_value(cls.VALID_FLAGS)
        return self

    @classmethod
    def none(cls) -> Self:
        """A factory method that creates an :class:`AutoModKeywordPresets` instance with everything disabled."""
        self = cls.__new__(cls)
        self.value = self.DEFAULT_VALUE
        return self

    @flag_value
    def profanity(self):
        """:class:`bool`: Returns ``True`` if the profanity preset is enabled
        (contains words that may be considered swearing or cursing).
        """
        return 1 << 1

    @flag_value
    def sexual_content(self):
        """:class:`bool`: Returns ``True`` if the sexual content preset is enabled
        (contains sexually explicit words).
        """
        return 1 << 2

    @flag_value
    def slurs(self):
        """:class:`bool`: Returns ``True`` if the slurs preset is enabled
        (contains insults or words that may be considered hate speech).
        """
        return 1 << 3


class MemberFlags(BaseFlags):
    """Wraps up Discord Member flags.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two MemberFlags instances are equal.
        .. describe:: x != y

            Checks if two MemberFlags instances are not equal.
        .. describe:: x <= y

            Checks if a MemberFlags instance is a subset of another MemberFlags instance.
        .. describe:: x >= y

            Checks if a MemberFlags instance is a superset of another MemberFlags instance.
        .. describe:: x < y

            Checks if a MemberFlags instance is a strict subset of another MemberFlags instance.
        .. describe:: x > y

            Checks if a MemberFlags instance is a strict superset of another MemberFlags instance.
        .. describe:: x | y, x |= y

            Returns a new MemberFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).
        .. describe:: x & y, x &= y

            Returns a new MemberFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).
        .. describe:: x ^ y, x ^= y

            Returns a new MemberFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).
        .. describe:: ~x

            Returns a new MemberFlags instance with all flags from x inverted.
        .. describe:: hash(x)

            Returns the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

        Additionally supported are a few operations on class attributes.

        .. describe:: MemberFlags.y | MemberFlags.z, MemberFlags(y=True) | MemberFlags.z

            Returns a MemberFlags instance with all provided flags enabled.

        .. describe:: ~MemberFlags.y

            Returns a MemberFlags instance with all flags except ``y`` inverted from their default value.

    .. versionadded:: 2.8

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self,
            *,
            automod_quarantined_username: bool = ...,
            bypasses_verification: bool = ...,
            completed_home_actions: bool = ...,
            completed_onboarding: bool = ...,
            did_rejoin: bool = ...,
            dm_settings_upsell_acknowledged: bool = ...,
            is_guest: bool = ...,
            started_home_actions: bool = ...,
            started_onboarding: bool = ...,
        ) -> None:
            ...

    @flag_value
    def did_rejoin(self):
        """:class:`bool`: Returns ``True`` if the member has left and rejoined the guild."""
        return 1 << 0

    @flag_value
    def completed_onboarding(self):
        """:class:`bool`: Returns ``True`` if the member has completed onboarding."""
        return 1 << 1

    @flag_value
    def bypasses_verification(self):
        """:class:`bool`: Returns ``True`` if the member is able to bypass guild verification requirements."""
        return 1 << 2

    @flag_value
    def started_onboarding(self):
        """:class:`bool`: Returns ``True`` if the member has started onboarding."""
        return 1 << 3

    @flag_value
    def is_guest(self):
        """:class:`bool`: Returns ``True`` if the member is a guest and can only access the voice channel they were invited to.

        .. versionadded:: 2.10
        """
        return 1 << 4

    @flag_value
    def started_home_actions(self):
        """:class:`bool`: Returns ``True`` if the member has started the Server Guide actions.

        .. versionadded:: 2.10
        """
        return 1 << 5

    @flag_value
    def completed_home_actions(self):
        """:class:`bool`: Returns ``True`` if the member has completed the Server Guide actions.

        .. versionadded:: 2.10
        """
        return 1 << 6

    @flag_value
    def automod_quarantined_username(self):
        """:class:`bool`: Returns ``True`` if the member's username, display name, or nickname is blocked by AutoMod.

        .. versionadded:: 2.10
        """
        return 1 << 7

    @flag_value
    def dm_settings_upsell_acknowledged(self):
        """:class:`bool`: Returns ``True`` if the member has dismissed the DM settings upsell.

        .. versionadded:: 2.10
        """
        return 1 << 9


class RoleFlags(BaseFlags):
    """Wraps up Discord Role flags.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two RoleFlags instances are equal.
        .. describe:: x != y

            Checks if two RoleFlags instances are not equal.
        .. describe:: x <= y

            Checks if a RoleFlags instance is a subset of another RoleFlags instance.
        .. describe:: x >= y

            Checks if a RoleFlags instance is a superset of another RoleFlags instance.
        .. describe:: x < y

            Checks if a RoleFlags instance is a strict subset of another RoleFlags instance.
        .. describe:: x > y

            Checks if a RoleFlags instance is a strict superset of another RoleFlags instance.
        .. describe:: x | y, x |= y

            Returns a new RoleFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).
        .. describe:: x & y, x &= y

            Returns a new RoleFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).
        .. describe:: x ^ y, x ^= y

            Returns a new RoleFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).
        .. describe:: ~x

            Returns a new RoleFlags instance with all flags from x inverted.
        .. describe:: hash(x)

            Returns the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

        Additionally supported are a few operations on class attributes.

        .. describe:: RoleFlags.y | RoleFlags.z, RoleFlags(y=True) | RoleFlags.z

            Returns a RoleFlags instance with all provided flags enabled.

        .. describe:: ~RoleFlags.y

            Returns a RoleFlags instance with all flags except ``y`` inverted from their default value.

    .. versionadded:: 2.10

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(self, *, in_prompt: bool = ...) -> None:
            ...

    @flag_value
    def in_prompt(self):
        """:class:`bool`: Returns ``True`` if the role can be selected by members in an onboarding prompt."""
        return 1 << 0


class AttachmentFlags(BaseFlags):
    """Wraps up Discord Attachment flags.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two AttachmentFlags instances are equal.
        .. describe:: x != y

            Checks if two AttachmentFlags instances are not equal.
        .. describe:: x <= y

            Checks if an AttachmentFlags instance is a subset of another AttachmentFlags instance.
        .. describe:: x >= y

            Checks if an AttachmentFlags instance is a superset of another AttachmentFlags instance.
        .. describe:: x < y

            Checks if an AttachmentFlags instance is a strict subset of another AttachmentFlags instance.
        .. describe:: x > y

            Checks if an AttachmentFlags instance is a strict superset of another AttachmentFlags instance.
        .. describe:: x | y, x |= y

            Returns a new AttachmentFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).
        .. describe:: x & y, x &= y

            Returns a new AttachmentFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).
        .. describe:: x ^ y, x ^= y

            Returns a new AttachmentFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).
        .. describe:: ~x

            Returns a new AttachmentFlags instance with all flags from x inverted.
        .. describe:: hash(x)

            Returns the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

        Additionally supported are a few operations on class attributes.

        .. describe:: AttachmentFlags.y | AttachmentFlags.z, AttachmentFlags(y=True) | AttachmentFlags.z

            Returns an AttachmentFlags instance with all provided flags enabled.

        .. describe:: ~AttachmentFlags.y

            Returns an AttachmentFlags instance with all flags except ``y`` inverted from their default value.

    .. versionadded:: 2.10

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(self, *, is_remix: bool = ...) -> None:
            ...

    @flag_value
    def is_remix(self):
        """:class:`bool`: Returns ``True`` if the attachment has been edited using the Remix feature."""
        return 1 << 2


class SKUFlags(BaseFlags):
    """Wraps up Discord SKU flags.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two SKUFlags instances are equal.
        .. describe:: x != y

            Checks if two SKUFlags instances are not equal.
        .. describe:: x <= y

            Checks if an SKUFlags instance is a subset of another SKUFlags instance.
        .. describe:: x >= y

            Checks if an SKUFlags instance is a superset of another SKUFlags instance.
        .. describe:: x < y

            Checks if an SKUFlags instance is a strict subset of another SKUFlags instance.
        .. describe:: x > y

            Checks if an SKUFlags instance is a strict superset of another SKUFlags instance.
        .. describe:: x | y, x |= y

            Returns a new SKUFlags instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).
        .. describe:: x & y, x &= y

            Returns a new SKUFlags instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).
        .. describe:: x ^ y, x ^= y

            Returns a new SKUFlags instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).
        .. describe:: ~x

            Returns a new SKUFlags instance with all flags from x inverted.
        .. describe:: hash(x)

            Returns the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

        Additionally supported are a few operations on class attributes.

        .. describe:: SKUFlags.y | SKUFlags.z, SKUFlags(y=True) | SKUFlags.z

            Returns an SKUFlags instance with all provided flags enabled.

        .. describe:: ~SKUFlags.y

            Returns an SKUFlags instance with all flags except ``y`` inverted from their default value.

    .. versionadded:: 2.10

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self,
            *,
            available: bool = ...,
            guild_subscription: bool = ...,
            user_subscription: bool = ...,
        ) -> None:
            ...

    @flag_value
    def available(self):
        """:class:`bool`: Returns ``True`` if the SKU can be purchased."""
        return 1 << 2

    @flag_value
    def guild_subscription(self):
        """:class:`bool`: Returns ``True`` if the SKU is an application subscription applied to a guild."""
        return 1 << 7

    @flag_value
    def user_subscription(self):
        """:class:`bool`: Returns ``True`` if the SKU is an application subscription applied to a user."""
        return 1 << 8


class ApplicationInstallTypes(ListBaseFlags):
    """Represents the location(s) in which an application or application command can be installed.

    See the :ddocs:`official documentation <resources/application#installation-context>` for more info.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two ApplicationInstallTypes instances are equal.
        .. describe:: x != y

            Checks if two ApplicationInstallTypes instances are not equal.
        .. describe:: x <= y

            Checks if an ApplicationInstallTypes instance is a subset of another ApplicationInstallTypes instance.
        .. describe:: x >= y

            Checks if an ApplicationInstallTypes instance is a superset of another ApplicationInstallTypes instance.
        .. describe:: x < y

            Checks if an ApplicationInstallTypes instance is a strict subset of another ApplicationInstallTypes instance.
        .. describe:: x > y

            Checks if an ApplicationInstallTypes instance is a strict superset of another ApplicationInstallTypes instance.
        .. describe:: x | y, x |= y

            Returns a new ApplicationInstallTypes instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).
        .. describe:: x & y, x &= y

            Returns a new ApplicationInstallTypes instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).
        .. describe:: x ^ y, x ^= y

            Returns a new ApplicationInstallTypes instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).
        .. describe:: ~x

            Returns a new ApplicationInstallTypes instance with all flags from x inverted.
        .. describe:: hash(x)

            Returns the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

        Additionally supported are a few operations on class attributes.

        .. describe:: ApplicationInstallTypes.y | ApplicationInstallTypes.z, ApplicationInstallTypes(y=True) | ApplicationInstallTypes.z

            Returns an ApplicationInstallTypes instance with all provided flags enabled.

        .. describe:: ~ApplicationInstallTypes.y

            Returns an ApplicationInstallTypes instance with all flags except ``y`` inverted from their default value.

    .. versionadded:: 2.10

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(self, *, guild: bool = ..., user: bool = ...) -> None:
            ...

    @classmethod
    def all(cls) -> Self:
        """A factory method that creates an :class:`ApplicationInstallTypes` instance with everything enabled."""
        self = cls.__new__(cls)
        self.value = all_flags_value(cls.VALID_FLAGS)
        return self

    @flag_value
    def guild(self):
        """:class:`bool`: Returns ``True`` if installable to guilds."""
        return 1 << 0

    @flag_value
    def user(self):
        """:class:`bool`: Returns ``True`` if installable to users."""
        return 1 << 1


class InteractionContextTypes(ListBaseFlags):
    """Represents the context(s) in which an application command can be used.

    See the :ddocs:`official documentation <interactions/application-commands#interaction-contexts>` for more info.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two InteractionContextTypes instances are equal.
        .. describe:: x != y

            Checks if two InteractionContextTypes instances are not equal.
        .. describe:: x <= y

            Checks if an InteractionContextTypes instance is a subset of another InteractionContextTypes instance.
        .. describe:: x >= y

            Checks if an InteractionContextTypes instance is a superset of another InteractionContextTypes instance.
        .. describe:: x < y

            Checks if an InteractionContextTypes instance is a strict subset of another InteractionContextTypes instance.
        .. describe:: x > y

            Checks if an InteractionContextTypes instance is a strict superset of another InteractionContextTypes instance.
        .. describe:: x | y, x |= y

            Returns a new InteractionContextTypes instance with all enabled flags from both x and y.
            (Using ``|=`` will update in place).
        .. describe:: x & y, x &= y

            Returns a new InteractionContextTypes instance with only flags enabled on both x and y.
            (Using ``&=`` will update in place).
        .. describe:: x ^ y, x ^= y

            Returns a new InteractionContextTypes instance with only flags enabled on one of x or y, but not both.
            (Using ``^=`` will update in place).
        .. describe:: ~x

            Returns a new InteractionContextTypes instance with all flags from x inverted.
        .. describe:: hash(x)

            Returns the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

        Additionally supported are a few operations on class attributes.

        .. describe:: InteractionContextTypes.y | InteractionContextTypes.z, InteractionContextTypes(y=True) | InteractionContextTypes.z

            Returns an InteractionContextTypes instance with all provided flags enabled.

        .. describe:: ~InteractionContextTypes.y

            Returns an InteractionContextTypes instance with all flags except ``y`` inverted from their default value.

    .. versionadded:: 2.10

    Attributes
    ----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    if TYPE_CHECKING:

        @_generated
        def __init__(
            self, *, bot_dm: bool = ..., guild: bool = ..., private_channel: bool = ...
        ) -> None:
            ...

    @classmethod
    def all(cls) -> Self:
        """A factory method that creates an :class:`InteractionContextTypes` instance with everything enabled."""
        self = cls.__new__(cls)
        self.value = all_flags_value(cls.VALID_FLAGS)
        return self

    @flag_value
    def guild(self):
        """:class:`bool`: Returns ``True`` if the command is usable in guilds."""
        return 1 << 0

    @flag_value
    def bot_dm(self):
        """:class:`bool`: Returns ``True`` if the command is usable in DMs with the bot."""
        return 1 << 1

    @flag_value
    def private_channel(self):
        """:class:`bool`: Returns ``True`` if the command is usable in DMs and group DMs with other users."""
        return 1 << 2



# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from disnake import utils
from disnake.app_commands import Option, SlashCommand
from disnake.enums import OptionType
from disnake.flags import ApplicationInstallTypes, InteractionContextTypes
from disnake.i18n import Localized
from disnake.interactions import ApplicationCommandInteraction
from disnake.permissions import Permissions

from .base_core import InvokableApplicationCommand, _get_overridden_method
from .errors import CommandError, CommandInvokeError
from .params import call_param_func, classify_autocompleter, expand_params

if TYPE_CHECKING:
    from disnake.app_commands import Choices
    from disnake.i18n import LocalizedOptional

    from .base_core import CommandCallback

MISSING = utils.MISSING

__all__ = ("InvokableSlashCommand", "SubCommandGroup", "SubCommand", "slash_command")


SlashCommandT = TypeVar("SlashCommandT", bound="InvokableSlashCommand")


def _autocomplete(
    self: Union[SubCommand, InvokableSlashCommand], option_name: str
) -> Callable[[Callable], Callable]:
    for option in self.body.options:
        if option.name == option_name:
            option.autocomplete = True
            break
    else:  # nobreak
        raise ValueError(f"Option '{option_name}' doesn't exist in '{self.qualified_name}'")

    def decorator(func: Callable) -> Callable:
        classify_autocompleter(func)
        self.autocompleters[option_name] = func
        return func

    return decorator


async def _call_autocompleter(
    self: Union[InvokableSlashCommand, SubCommand],
    param: str,
    inter: ApplicationCommandInteraction,
    user_input: str,
) -> Optional[Choices]:
    autocomp = self.autocompleters.get(param)
    if autocomp is None:
        return None
    if not callable(autocomp):
        return autocomp

    try:
        requires_cog_param = autocomp.__has_cog_param__
    except AttributeError:
        requires_cog_param = False

    cog = self.root_parent.cog if isinstance(self, SubCommand) else self.cog
    filled = inter.filled_options
    del filled[inter.data.focused_option.name]

    try:
        if requires_cog_param:
            choices = autocomp(cog, inter, user_input, **filled)
        else:
            choices = autocomp(inter, user_input, **filled)
    except TypeError:
        if requires_cog_param:
            choices = autocomp(cog, inter, user_input)
        else:
            choices = autocomp(inter, user_input)

    if inspect.isawaitable(choices):
        return await choices
    return choices


_INVALID_SUB_KWARGS = frozenset(
    {"dm_permission", "default_member_permissions", "install_types", "contexts"}
)


# this is just a helpful message for users trying to set specific
# top-level-only fields on subcommands or groups
def _check_invalid_sub_kwargs(func: CommandCallback, kwargs: Dict[str, Any]) -> None:
    invalid_keys = kwargs.keys() & _INVALID_SUB_KWARGS

    for decorator_key in [
        "__default_member_permissions__",
        "__install_types__",
        "__contexts__",
    ]:
        if hasattr(func, decorator_key):
            invalid_keys.add(decorator_key.strip("_"))

    if invalid_keys:
        msg = f"Cannot set {utils.humanize_list(list(invalid_keys), 'or')} on subcommands or subcommand groups"
        raise TypeError(msg)


class SubCommandGroup(InvokableApplicationCommand):
    """A class that implements the protocol for a bot slash command group.

    These are not created manually, instead they are created via the
    decorator or functional interface.

    Attributes
    ----------
    name: :class:`str`
        The name of the group.
    qualified_name: :class:`str`
        The full command name, including parent names in the case of slash subcommands or groups.
        For example, the qualified name for ``/one two three`` would be ``one two three``.
    parent: :class:`InvokableSlashCommand`
        The parent command this group belongs to.

        .. versionadded:: 2.6
    option: :class:`.Option`
        API representation of this subcommand.
    callback: :ref:`coroutine <coroutine>`
        The coroutine that is executed when the command group is invoked.
    cog: Optional[:class:`Cog`]
        The cog that this group belongs to. ``None`` if there isn't one.
    checks: List[Callable[[:class:`.ApplicationCommandInteraction`], :class:`bool`]]
        A list of predicates that verifies if the group could be executed
        with the given :class:`.ApplicationCommandInteraction` as the sole parameter. If an exception
        is necessary to be thrown to signal failure, then one inherited from
        :exc:`.CommandError` should be used. Note that if the checks fail then
        :exc:`.CheckFailure` exception is raised to the :func:`.on_slash_command_error`
        event.
    extras: Dict[:class:`str`, Any]
        A dict of user provided extras to attach to the subcommand group.

        .. note::
            This object may be copied by the library.

        .. versionadded:: 2.5
    """

    def __init__(
        self,
        func: CommandCallback,
        parent: InvokableSlashCommand,
        *,
        name: LocalizedOptional = None,
        **kwargs: Any,
    ) -> None:
        name_loc = Localized._cast(name, False)
        super().__init__(func, name=name_loc.string, **kwargs)
        self.parent: InvokableSlashCommand = parent
        self.children: Dict[str, SubCommand] = {}
        self.option = Option(
            name=name_loc._upgrade(self.name),
            description="-",
            type=OptionType.sub_command_group,
            options=[],
        )
        self.qualified_name: str = f"{parent.qualified_name} {self.name}"

        _check_invalid_sub_kwargs(func, kwargs)

    @property
    def root_parent(self) -> InvokableSlashCommand:
        """:class:`InvokableSlashCommand`: Returns the slash command containing this group.
        This is mainly for consistency with :class:`SubCommand`, and is equivalent to :attr:`parent`.

        .. versionadded:: 2.6
        """
        return self.parent

    @property
    def parents(self) -> Tuple[InvokableSlashCommand]:
        """Tuple[:class:`InvokableSlashCommand`]: Returns all parents of this group.

        .. versionadded:: 2.6
        """
        return (self.parent,)

    @property
    def body(self) -> Option:
        return self.option

    def sub_command(
        self,
        name: LocalizedOptional = None,
        description: LocalizedOptional = None,
        options: Optional[list] = None,
        connectors: Optional[dict] = None,
        extras: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Callable[[CommandCallback], SubCommand]:
        """A decorator that creates a subcommand in the subcommand group.
        Parameters are the same as in :class:`InvokableSlashCommand.sub_command`

        Returns
        -------
        Callable[..., :class:`SubCommand`]
            A decorator that converts the provided method into a SubCommand, adds it to the bot, then returns it.
        """

        def decorator(func: CommandCallback) -> SubCommand:
            new_func = SubCommand(
                func,
                self,
                name=name,
                description=description,
                options=options,
                connectors=connectors,
                extras=extras,
                **kwargs,
            )
            self.children[new_func.name] = new_func
            self.option.options.append(new_func.option)
            return new_func

        return decorator


class SubCommand(InvokableApplicationCommand):
    """A class that implements the protocol for a bot slash subcommand.

    These are not created manually, instead they are created via the
    decorator or functional interface.

    Attributes
    ----------
    name: :class:`str`
        The name of the subcommand.
    qualified_name: :class:`str`
        The full command name, including parent names in the case of slash subcommands or groups.
        For example, the qualified name for ``/one two three`` would be ``one two three``.
    parent: Union[:class:`InvokableSlashCommand`, :class:`SubCommandGroup`]
        The parent command or group this subcommand belongs to.

        .. versionadded:: 2.6
    option: :class:`.Option`
        API representation of this subcommand.
    callback: :ref:`coroutine <coroutine>`
        The coroutine that is executed when the subcommand is called.
    cog: Optional[:class:`Cog`]
        The cog that this subcommand belongs to. ``None`` if there isn't one.
    checks: List[Callable[[:class:`.ApplicationCommandInteraction`], :class:`bool`]]
        A list of predicates that verifies if the subcommand could be executed
        with the given :class:`.ApplicationCommandInteraction` as the sole parameter. If an exception
        is necessary to be thrown to signal failure, then one inherited from
        :exc:`.CommandError` should be used. Note that if the checks fail then
        :exc:`.CheckFailure` exception is raised to the :func:`.on_slash_command_error`
        event.
    connectors: Dict[:class:`str`, :class:`str`]
        A mapping of option names to function parameter names, mainly for internal processes.
    extras: Dict[:class:`str`, Any]
        A dict of user provided extras to attach to the subcommand.

        .. note::
            This object may be copied by the library.

        .. versionadded:: 2.5
    """

    def __init__(
        self,
        func: CommandCallback,
        parent: Union[InvokableSlashCommand, SubCommandGroup],
        *,
        name: LocalizedOptional = None,
        description: LocalizedOptional = None,
        options: Optional[list] = None,
        connectors: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> None:
        name_loc = Localized._cast(name, False)
        super().__init__(func, name=name_loc.string, **kwargs)
        self.parent: Union[InvokableSlashCommand, SubCommandGroup] = parent
        self.connectors: Dict[str, str] = connectors or {}
        self.autocompleters: Dict[str, Any] = kwargs.get("autocompleters", {})

        if options is None:
            options = expand_params(self)

        self.docstring = utils.parse_docstring(func)
        desc_loc = Localized._cast(description, False)

        self.option = Option(
            name=name_loc._upgrade(self.name, key=self.docstring["localization_key_name"]),
            description=desc_loc._upgrade(
                self.docstring["description"] or "-", key=self.docstring["localization_key_desc"]
            ),
            type=OptionType.sub_command,
            options=options,
        )
        self.qualified_name = f"{parent.qualified_name} {self.name}"

        _check_invalid_sub_kwargs(func, kwargs)

    @property
    def root_parent(self) -> InvokableSlashCommand:
        """:class:`InvokableSlashCommand`: Returns the top-level slash command containing this subcommand,
        even if the parent is a :class:`SubCommandGroup`.

        .. versionadded:: 2.6
        """
        return self.parent.parent if isinstance(self.parent, SubCommandGroup) else self.parent

    @property
    def parents(
        self,
    ) -> Union[Tuple[InvokableSlashCommand], Tuple[SubCommandGroup, InvokableSlashCommand]]:
        """Union[Tuple[:class:`InvokableSlashCommand`], Tuple[:class:`SubCommandGroup`, :class:`InvokableSlashCommand`]]:
        Returns all parents of this subcommand.

        For example, the parents of the ``c`` subcommand in ``/a b c`` are ``(b, a)``.

        .. versionadded:: 2.6
        """
        # here I'm not using 'self.parent.parents + (self.parent,)' because it causes typing issues
        if isinstance(self.parent, SubCommandGroup):
            return (self.parent, self.parent.parent)
        return (self.parent,)

    @property
    def description(self) -> str:
        """:class:`str`: The slash sub command's description. Shorthand for :attr:`self.body.description <.Option.description>`."""
        return self.body.description

    @property
    def body(self) -> Option:
        """:class:`.Option`: The API representation for this slash sub command. Shorthand for :attr:`.SubCommand.option`"""
        return self.option

    async def _call_autocompleter(
        self, param: str, inter: ApplicationCommandInteraction, user_input: str
    ) -> Optional[Choices]:
        return await _call_autocompleter(self, param, inter, user_input)

    async def invoke(self, inter: ApplicationCommandInteraction, *args: Any, **kwargs: Any) -> None:
        for k, v in self.connectors.items():
            if k in kwargs:
                kwargs[v] = kwargs.pop(k)

        await self.prepare(inter)

        try:
            await call_param_func(self.callback, inter, self.cog, **kwargs)
        except CommandError:
            inter.command_failed = True
            raise
        except asyncio.CancelledError:
            inter.command_failed = True
            return
        except Exception as exc:
            inter.command_failed = True
            raise CommandInvokeError(exc) from exc
        finally:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(inter)  # type: ignore

            await self.call_after_hooks(inter)

    def autocomplete(self, option_name: str) -> Callable[[Callable], Callable]:
        """A decorator that registers an autocomplete function for the specified option.

        Parameters
        ----------
        option_name: :class:`str`
            The name of the slash command option.
        """
        return _autocomplete(self, option_name)


class InvokableSlashCommand(InvokableApplicationCommand):
    """A class that implements the protocol for a bot slash command.

    These are not created manually, instead they are created via the
    decorator or functional interface.

    Attributes
    ----------
    name: :class:`str`
        The name of the command.
    qualified_name: :class:`str`
        The full command name, including parent names in the case of slash subcommands or groups.
        For example, the qualified name for ``/one two three`` would be ``one two three``.
    body: :class:`.SlashCommand`
        An object being registered in the API.
    callback: :ref:`coroutine <coroutine>`
        The coroutine that is executed when the command is called.
    cog: Optional[:class:`Cog`]
        The cog that this command belongs to. ``None`` if there isn't one.
    checks: List[Callable[[:class:`.ApplicationCommandInteraction`], :class:`bool`]]
        A list of predicates that verifies if the command could be executed
        with the given :class:`.ApplicationCommandInteraction` as the sole parameter. If an exception
        is necessary to be thrown to signal failure, then one inherited from
        :exc:`.CommandError` should be used. Note that if the checks fail then
        :exc:`.CheckFailure` exception is raised to the :func:`.on_slash_command_error`
        event.
    guild_ids: Optional[Tuple[:class:`int`, ...]]
        The list of IDs of the guilds where the command is synced. ``None`` if this command is global.
    connectors: Dict[:class:`str`, :class:`str`]
        A mapping of option names to function parameter names, mainly for internal processes.
    auto_sync: :class:`bool`
        Whether to automatically register the command.
    extras: Dict[:class:`str`, Any]
        A dict of user provided extras to attach to the command.

        .. note::
            This object may be copied by the library.

        .. versionadded:: 2.5

    parent: ``None``
        This exists for consistency with :class:`SubCommand` and :class:`SubCommandGroup`. Always ``None``.

        .. versionadded:: 2.6
    """

    def __init__(
        self,
        func: CommandCallback,
        *,
        name: LocalizedOptional = None,
        description: LocalizedOptional = None,
        options: Optional[List[Option]] = None,
        dm_permission: Optional[bool] = None,  # deprecated
        default_member_permissions: Optional[Union[Permissions, int]] = None,
        nsfw: Optional[bool] = None,
        install_types: Optional[ApplicationInstallTypes] = None,
        contexts: Optional[InteractionContextTypes] = None,
        guild_ids: Optional[Sequence[int]] = None,
        connectors: Optional[Dict[str, str]] = None,
        auto_sync: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        name_loc = Localized._cast(name, False)
        super().__init__(func, name=name_loc.string, **kwargs)
        self.parent = None
        self.connectors: Dict[str, str] = connectors or {}
        self.children: Dict[str, Union[SubCommand, SubCommandGroup]] = {}
        self.auto_sync: bool = True if auto_sync is None else auto_sync
        self.guild_ids: Optional[Tuple[int, ...]] = None if guild_ids is None else tuple(guild_ids)
        self.autocompleters: Dict[str, Any] = kwargs.get("autocompleters", {})

        if options is None:
            options = expand_params(self)

        self.docstring = utils.parse_docstring(func)
        desc_loc = Localized._cast(description, False)

        try:
            default_member_permissions = func.__default_member_permissions__
        except AttributeError:
            pass
        try:
            install_types = func.__install_types__
        except AttributeError:
            pass
        try:
            contexts = func.__contexts__
        except AttributeError:
            pass

        self.body: SlashCommand = SlashCommand(
            name=name_loc._upgrade(self.name, key=self.docstring["localization_key_name"]),
            description=desc_loc._upgrade(
                self.docstring["description"] or "-", key=self.docstring["localization_key_desc"]
            ),
            options=options or [],
            dm_permission=dm_permission,
            default_member_permissions=default_member_permissions,
            nsfw=nsfw,
            install_types=install_types,
            contexts=contexts,
        )

        self._apply_guild_only()

    @property
    def root_parent(self) -> None:
        """``None``: This is for consistency with :class:`SubCommand` and :class:`SubCommandGroup`.

        .. versionadded:: 2.6
        """
        return None

    @property
    def parents(self) -> Tuple[()]:
        """Tuple[()]: This is mainly for consistency with :class:`SubCommand`, and is equivalent to an empty tuple.

        .. versionadded:: 2.6
        """
        return ()

    def _ensure_assignment_on_copy(self, other: SlashCommandT) -> SlashCommandT:
        super()._ensure_assignment_on_copy(other)
        if self.connectors != other.connectors:
            other.connectors = self.connectors.copy()
        if self.autocompleters != other.autocompleters:
            other.autocompleters = self.autocompleters.copy()
        if self.children != other.children:
            other.children = self.children.copy()
            # update parents...
            for child in other.children.values():
                child.parent = other
        if self.description != other.description and "description" not in other.__original_kwargs__:
            # Allows overriding the default description cog-wide.
            other.body.description = self.description
        if self.options != other.options:
            other.body.options = self.options
        return other

    @property
    def description(self) -> str:
        """:class:`str`: The slash command's description. Shorthand for :attr:`self.body.description <.SlashCommand.description>`."""
        return self.body.description

    @property
    def options(self) -> List[Option]:
        """List[:class:`.Option`]: The list of options the slash command has. Shorthand for :attr:`self.body.options <.SlashCommand.options>`."""
        return self.body.options

    def sub_command(
        self,
        name: LocalizedOptional = None,
        description: LocalizedOptional = None,
        options: Optional[list] = None,
        connectors: Optional[dict] = None,
        extras: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Callable[[CommandCallback], SubCommand]:
        """A decorator that creates a subcommand under the base command.

        Parameters
        ----------
        name: Optional[Union[:class:`str`, :class:`.Localized`]]
            The name of the subcommand (defaults to function name).

            .. versionchanged:: 2.5
                Added support for localizations.

        description: Optional[Union[:class:`str`, :class:`.Localized`]]
            The description of the subcommand.

            .. versionchanged:: 2.5
                Added support for localizations.

        options: List[:class:`.Option`]
            the options of the subcommand for registration in API
        connectors: Dict[:class:`str`, :class:`str`]
            which function param states for each option. If the name
            of an option already matches the corresponding function param,
            you don't have to specify the connectors. Connectors template:
            ``{"option-name": "param_name", ...}``
        extras: Dict[:class:`str`, Any]
            A dict of user provided extras to attach to the subcommand.

            .. note::
                This object may be copied by the library.

            .. versionadded:: 2.5

        Returns
        -------
        Callable[..., :class:`SubCommand`]
            A decorator that converts the provided method into a :class:`SubCommand`, adds it to the bot, then returns it.
        """

        def decorator(func: CommandCallback) -> SubCommand:
            if len(self.children) == 0 and len(self.body.options) > 0:
                self.body.options = []
            new_func = SubCommand(
                func,
                self,
                name=name,
                description=description,
                options=options,
                connectors=connectors,
                extras=extras,
                **kwargs,
            )
            self.children[new_func.name] = new_func
            self.body.options.append(new_func.option)
            return new_func

        return decorator

    def sub_command_group(
        self,
        name: LocalizedOptional = None,
        extras: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Callable[[CommandCallback], SubCommandGroup]:
        """A decorator that creates a subcommand group under the base command.

        Parameters
        ----------
        name: Optional[Union[:class:`str`, :class:`.Localized`]]
            The name of the subcommand group (defaults to function name).

            .. versionchanged:: 2.5
                Added support for localizations.
        extras: Dict[:class:`str`, Any]
            A dict of user provided extras to attach to the subcommand group.

            .. note::
                This object may be copied by the library.

            .. versionadded:: 2.5

        Returns
        -------
        Callable[..., :class:`SubCommandGroup`]
            A decorator that converts the provided method into a :class:`SubCommandGroup`, adds it to the bot, then returns it.
        """

        def decorator(func: CommandCallback) -> SubCommandGroup:
            if len(self.children) == 0 and len(self.body.options) > 0:
                self.body.options = []
            new_func = SubCommandGroup(
                func,
                self,
                name=name,
                extras=extras,
                **kwargs,
            )
            self.children[new_func.name] = new_func
            self.body.options.append(new_func.option)
            return new_func

        return decorator

    def autocomplete(self, option_name: str) -> Callable[[Callable], Callable]:
        """A decorator that registers an autocomplete function for the specified option.

        Parameters
        ----------
        option_name: :class:`str`
            The name of the slash command option.
        """
        return _autocomplete(self, option_name)

    async def _call_external_error_handlers(
        self, inter: ApplicationCommandInteraction, error: CommandError
    ) -> None:
        stop_propagation = False
        cog = self.cog
        try:
            if cog is not None:
                local = _get_overridden_method(cog.cog_slash_command_error)
                if local is not None:
                    stop_propagation = await local(inter, error)
                    # User has an option to cancel the global error handler by returning True
        finally:
            if stop_propagation:
                return  # noqa: B012
            inter.bot.dispatch("slash_command_error", inter, error)

    async def _call_autocompleter(
        self, param: str, inter: ApplicationCommandInteraction, user_input: str
    ) -> Optional[Choices]:
        return await _call_autocompleter(self, param, inter, user_input)

    async def _call_relevant_autocompleter(self, inter: ApplicationCommandInteraction) -> None:
        chain, _ = inter.data._get_chain_and_kwargs()

        if len(chain) == 0:
            subcmd = None
        elif len(chain) == 1:
            subcmd = self.children.get(chain[0])
        elif len(chain) == 2:
            group = self.children.get(chain[0])
            if not isinstance(group, SubCommandGroup):
                raise AssertionError("the first subcommand is not a SubCommandGroup instance")
            subcmd = group.children.get(chain[1])
        else:
            raise ValueError("Command chain is too long")

        focused_option = inter.data.focused_option

        if subcmd is None or isinstance(subcmd, SubCommandGroup):
            call_autocompleter = self._call_autocompleter
        else:
            call_autocompleter = subcmd._call_autocompleter

        choices = await call_autocompleter(focused_option.name, inter, focused_option.value)

        if choices is not None:
            await inter.response.autocomplete(choices=choices)

    async def invoke_children(self, inter: ApplicationCommandInteraction) -> None:
        chain, kwargs = inter.data._get_chain_and_kwargs()

        if len(chain) == 0:
            group = None
            subcmd = None
        elif len(chain) == 1:
            group = None
            subcmd = self.children.get(chain[0])
        elif len(chain) == 2:
            group = self.children.get(chain[0])
            if not isinstance(group, SubCommandGroup):
                raise AssertionError("the first subcommand is not a SubCommandGroup instance")
            subcmd = group.children.get(chain[1])
        else:
            raise ValueError("Command chain is too long")

        if group is not None:
            try:
                await group.invoke(inter)
            except CommandError as exc:
                if not await group._call_local_error_handler(inter, exc):
                    raise

        if subcmd is not None:
            try:
                await subcmd.invoke(inter, **kwargs)
            except CommandError as exc:
                if not await subcmd._call_local_error_handler(inter, exc):
                    raise

    async def invoke(self, inter: ApplicationCommandInteraction) -> None:
        await self.prepare(inter)

        try:
            if len(self.children) > 0:
                await self(inter)
                await self.invoke_children(inter)
            else:
                kwargs = inter.filled_options
                for k, v in self.connectors.items():
                    if k in kwargs:
                        kwargs[v] = kwargs.pop(k)
                await call_param_func(self.callback, inter, self.cog, **kwargs)
        except CommandError:
            inter.command_failed = True
            raise
        except asyncio.CancelledError:
            inter.command_failed = True
            return
        except Exception as exc:
            inter.command_failed = True
            raise CommandInvokeError(exc) from exc
        finally:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(inter)  # type: ignore

            await self.call_after_hooks(inter)


def slash_command(
    *,
    name: LocalizedOptional = None,
    description: LocalizedOptional = None,
    dm_permission: Optional[bool] = None,  # deprecated
    default_member_permissions: Optional[Union[Permissions, int]] = None,
    nsfw: Optional[bool] = None,
    install_types: Optional[ApplicationInstallTypes] = None,
    contexts: Optional[InteractionContextTypes] = None,
    options: Optional[List[Option]] = None,
    guild_ids: Optional[Sequence[int]] = None,
    connectors: Optional[Dict[str, str]] = None,
    auto_sync: Optional[bool] = None,
    extras: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> Callable[[CommandCallback], InvokableSlashCommand]:
    """A decorator that builds a slash command.

    Parameters
    ----------
    auto_sync: :class:`bool`
        Whether to automatically register the command. Defaults to ``True``.
    name: Optional[Union[:class:`str`, :class:`.Localized`]]
        The name of the slash command (defaults to function name).

        .. versionchanged:: 2.5
            Added support for localizations.

    description: Optional[Union[:class:`str`, :class:`.Localized`]]
        The description of the slash command. It will be visible in Discord.

        .. versionchanged:: 2.5
            Added support for localizations.

    nsfw: :class:`bool`
        Whether this command is :ddocs:`age-restricted <interactions/application-commands#agerestricted-commands>`.
        Defaults to ``False``.

        .. versionadded:: 2.8

    install_types: Optional[:class:`.ApplicationInstallTypes`]
        The installation types where the command is available.
        Defaults to :attr:`.ApplicationInstallTypes.guild` only.
        Only available for global commands.

        See :ref:`app_command_contexts` for details.

        .. versionadded:: 2.10

    contexts: Optional[:class:`.InteractionContextTypes`]
        The interaction contexts where the command can be used.
        Only available for global commands.

        See :ref:`app_command_contexts` for details.

        .. versionadded:: 2.10

    options: List[:class:`.Option`]
        The list of slash command options. The options will be visible in Discord.
        This is the old way of specifying options. Consider using :ref:`param_syntax` instead.
    dm_permission: :class:`bool`
        Whether this command can be used in DMs.
        Defaults to ``True``.

        .. deprecated:: 2.10
            Use ``contexts`` instead.
            This is equivalent to the :attr:`.InteractionContextTypes.bot_dm` flag.

    default_member_permissions: Optional[Union[:class:`.Permissions`, :class:`int`]]
        The default required permissions for this command.
        See :attr:`.ApplicationCommand.default_member_permissions` for details.

        .. versionadded:: 2.5

    guild_ids: List[:class:`int`]
        If specified, the client will register the command in these guilds.
        Otherwise, this command will be registered globally.
    connectors: Dict[:class:`str`, :class:`str`]
        Binds function names to option names. If the name
        of an option already matches the corresponding function param,
        you don't have to specify the connectors. Connectors template:
        ``{"option-name": "param_name", ...}``.
        If you're using :ref:`param_syntax`, you don't need to specify this.
    extras: Dict[:class:`str`, Any]
        A dict of user provided extras to attach to the command.

        .. note::
            This object may be copied by the library.

        .. versionadded:: 2.5

    Returns
    -------
    Callable[..., :class:`InvokableSlashCommand`]
        A decorator that converts the provided method into an InvokableSlashCommand and returns it.
    """

    def decorator(func: CommandCallback) -> InvokableSlashCommand:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f"<{func.__qualname__}> must be a coroutine function")
        if hasattr(func, "__command_flag__"):
            raise TypeError("Callback is already a command.")
        if guild_ids and not all(isinstance(guild_id, int) for guild_id in guild_ids):
            raise ValueError("guild_ids must be a sequence of int.")
        return InvokableSlashCommand(
            func,
            name=name,
            description=description,
            options=options,
            dm_permission=dm_permission,
            default_member_permissions=default_member_permissions,
            nsfw=nsfw,
            install_types=install_types,
            contexts=contexts,
            guild_ids=guild_ids,
            connectors=connectors,
            auto_sync=auto_sync,
            extras=extras,
            **kwargs,
        )

    return decorator



# SPDX-License-Identifier: MIT

import pytest

from disnake.http import HTTPClient


@pytest.mark.parametrize(
    ("url", "encoding", "zlib", "expected"),
    [
        (
            "wss://gateway.discord.com",
            "json",
            False,
            "wss://gateway.discord.com/?v=10&encoding=json",
        ),
        (
            "wss://gateway.discord.com",
            "json",
            True,
            "wss://gateway.discord.com/?v=10&encoding=json&compress=zlib-stream",
        ),
        # should overwrite existing args if needed
        (
            "wss://gateway.discord.com/?v=42&encoding=etf&v=1111",
            "json",
            True,
            "wss://gateway.discord.com/?v=10&encoding=json&compress=zlib-stream",
        ),
        # should keep other args intact
        (
            "wss://gateway.discord.com/?v=42&stuff=things&a=b",
            "json",
            True,
            "wss://gateway.discord.com/?v=10&stuff=things&a=b&encoding=json&compress=zlib-stream",
        ),
        # should remove compression if set to false
        (
            "wss://gateway.discord.com/?v=10&compress=zlib-stream",
            "json",
            False,
            "wss://gateway.discord.com/?v=10&encoding=json",
        ),
    ],
)
def test_format_gateway_url(url: str, encoding: str, zlib: bool, expected: str) -> None:
    assert HTTPClient._format_gateway_url(url, encoding=encoding, zlib=zlib) == expected



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .types.voice import VoiceRegion as VoiceRegionPayload


__all__ = ("VoiceRegion",)


class VoiceRegion:
    """Represents a Discord voice region.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two :class:`VoiceRegion`\\s are equal.

            .. versionadded:: 2.9

        .. describe:: x != y

            Checks if two :class:`VoiceRegion`\\s are not equal.

            .. versionadded:: 2.9

        .. describe:: str(x)

            Returns the voice region's ID.

    .. versionadded:: 2.5

    Attributes
    ----------
    id: :class:`str`
        Unique ID for the region.
    name: :class:`str`
        The name of the region.
    optimal: :class:`bool`
        True for a single region that is closest to your client.
    deprecated: :class:`bool`
        Whether this is a deprecated voice region (avoid switching to these).
    custom: :class:`bool`
        Whether this is a custom voice region (used for events/etc).
    """

    __slots__ = (
        "id",
        "name",
        "optimal",
        "deprecated",
        "custom",
    )

    def __init__(self, *, data: VoiceRegionPayload) -> None:
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.deprecated: bool = data.get("deprecated", False)
        self.optimal: bool = data.get("optimal", False)
        self.custom: bool = data.get("custom", False)

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        return f"<VoiceRegion id={self.id!r} name={self.name!r} optimal={self.optimal!r}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, VoiceRegion) and self.id == other.id
